@@ -4,10 +4,10 @@
 
 #include "config.h"
 
-#include <glib.h>
-
 #ifdef ENABLE_DEBUG
 
+#include <glib.h>
+
 G_BEGIN_DECLS
 
 typedef enum
@@ -23,6 +23,7 @@ typedef enum
   GABBLE_DEBUG_CONNECTION    = 1 << 8,
   GABBLE_DEBUG_IM            = 1 << 9,
   GABBLE_DEBUG_PERSIST       = 1 << 10,
+  GABBLE_DEBUG_VCARD         = 1 << 11,
 } GabbleDebugFlags;
 
 void gabble_debug_set_flags_from_env ();
@@ -31,6 +32,8 @@ gboolean gabble_debug_flag_is_set (GabbleDebugFlags flag);
 void gabble_debug (GabbleDebugFlags flag, const gchar *format, ...)
     G_GNUC_PRINTF (2, 3);
 
+G_END_DECLS
+
 #ifdef DEBUG_FLAG
 
 #define DEBUG(format, ...) \
@@ -61,7 +64,5 @@ G_STMT_START { \
 
 #endif /* ENABLE_DEBUG */
 
-G_END_DECLS
-
-#endif
+#endif /* __DEBUG_H__ */
 
