@@ -29,17 +29,15 @@
 #include "disco.h"
 #include "gabble-connection.h"
 #include "handles.h"
-#include "handle-set.h"
-#include "telepathy-constants.h"
-#include "telepathy-interfaces.h"
-#include "telepathy-helpers.h"
-#include "tp-channel-iface.h"
+#include <telepathy-glib/dbus.h>
+#include <telepathy-glib/enums.h>
+#include <telepathy-glib/interfaces.h>
+#include <telepathy-glib/channel-iface.h>
+#include <telepathy-glib/svc-channel.h>
 #include "namespaces.h"
 #include "util.h"
 
 #include "gabble-roomlist-channel.h"
-#include "gabble-roomlist-channel-glue.h"
-#include "gabble-roomlist-channel-signals-marshal.h"
 
 #define TP_TYPE_ROOM_STRUCT (dbus_g_type_get_struct ("GValueArray", \
       G_TYPE_UINT, \
@@ -50,21 +48,16 @@
 #define TP_TYPE_ROOM_LIST (dbus_g_type_get_collection ("GPtrArray", \
       TP_TYPE_ROOM_STRUCT))
 
-#define DISCO_PIPELINE_SIZE 10
+static void channel_iface_init (gpointer, gpointer);
+static void roomlist_iface_init (gpointer, gpointer);
 
 G_DEFINE_TYPE_WITH_CODE (GabbleRoomlistChannel, gabble_roomlist_channel,
-    G_TYPE_OBJECT, G_IMPLEMENT_INTERFACE (TP_TYPE_CHANNEL_IFACE, NULL));
-
-/* signal enum */
-enum
-{
-    CLOSED,
-    GOT_ROOMS,
-    LISTING_ROOMS,
-    LAST_SIGNAL
-};
-
-static guint signals[LAST_SIGNAL] = {0};
+    G_TYPE_OBJECT,
+    G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_CHANNEL, channel_iface_init);
+    G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_CHANNEL_TYPE_ROOM_LIST,
+      roomlist_iface_init);
+    G_IMPLEMENT_INTERFACE (TP_TYPE_CHANNEL_IFACE, NULL)
+    );
 
 /* properties */
 enum
@@ -90,9 +83,11 @@ struct _GabbleRoomlistChannelPrivate
   gboolean closed;
   gboolean listing;
 
-  GPtrArray *disco_pipeline;
-  GHashTable *remaining_rooms;
-  GabbleHandleSet *signalled_rooms;
+  gpointer disco_pipeline;
+  TpHandleSet *signalled_rooms;
+
+  GPtrArray *pending_room_signals;
+  guint timer_source_id;
 
   gboolean dispose_has_run;
 };
@@ -100,6 +95,10 @@ struct _GabbleRoomlistChannelPrivate
 #define GABBLE_ROOMLIST_CHANNEL_GET_PRIVATE(obj) \
     ((GabbleRoomlistChannelPrivate *)obj->priv)
 
+#define ROOM_SIGNAL_INTERVAL 300
+
+static gboolean emit_room_signal (gpointer data);
+
 static void
 gabble_roomlist_channel_init (GabbleRoomlistChannel *self)
 {
@@ -108,9 +107,7 @@ gabble_roomlist_channel_init (GabbleRoomlistChannel *self)
 
   self->priv = priv;
 
-  priv->disco_pipeline = g_ptr_array_sized_new (DISCO_PIPELINE_SIZE);
-  priv->remaining_rooms = g_hash_table_new_full (g_str_hash, g_str_equal,
-      g_free, NULL);
+  priv->pending_room_signals = g_ptr_array_new ();
 }
 
 
@@ -174,7 +171,8 @@ gabble_roomlist_channel_set_property (GObject     *object,
 {
   GabbleRoomlistChannel *chan = GABBLE_ROOMLIST_CHANNEL (object);
   GabbleRoomlistChannelPrivate *priv = GABBLE_ROOMLIST_CHANNEL_GET_PRIVATE (chan);
-  GabbleHandleSet *new_signalled_rooms;
+  TpBaseConnection *conn = (TpBaseConnection *)priv->conn;
+  TpHandleSet *new_signalled_rooms;
 
   switch (property_id) {
     case PROP_OBJECT_PATH:
@@ -185,17 +183,24 @@ gabble_roomlist_channel_set_property (GObject     *object,
       /* this property is writable in the interface, but not actually
        * meaningfully changable on this channel, so we do nothing */
       break;
+    case PROP_HANDLE_TYPE:
+      /* this property is writable in the interface, but not actually
+       * meaningfully changable on this channel, so we do nothing */
+      break;
     case PROP_CONNECTION:
       priv->conn = g_value_get_object (value);
-      new_signalled_rooms = handle_set_new (priv->conn->handles, TP_HANDLE_TYPE_ROOM);
-      if (priv->signalled_rooms)
+      conn = (TpBaseConnection *)priv->conn;
+
+      new_signalled_rooms = tp_handle_set_new (
+          conn->handles[TP_HANDLE_TYPE_ROOM]);
+      if (priv->signalled_rooms != NULL)
         {
-          const GIntSet *add;
-          GIntSet *tmp;
-          add = handle_set_peek (priv->signalled_rooms);
-          tmp = handle_set_update (new_signalled_rooms, add);
-          handle_set_destroy (priv->signalled_rooms);
-          g_intset_destroy (tmp);
+          const TpIntSet *add;
+          TpIntSet *tmp;
+          add = tp_handle_set_peek (priv->signalled_rooms);
+          tmp = tp_handle_set_update (new_signalled_rooms, add);
+          tp_handle_set_destroy (priv->signalled_rooms);
+          tp_intset_destroy (tmp);
         }
       priv->signalled_rooms = new_signalled_rooms;
       break;
@@ -254,54 +259,9 @@ gabble_roomlist_channel_class_init (GabbleRoomlistChannelClass *gabble_roomlist_
                                     G_PARAM_STATIC_BLURB);
   g_object_class_install_property (object_class, PROP_CONFERENCE_SERVER,
                                    param_spec);
-
-  signals[CLOSED] =
-    g_signal_new ("closed",
-                  G_OBJECT_CLASS_TYPE (gabble_roomlist_channel_class),
-                  G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
-                  0,
-                  NULL, NULL,
-                  g_cclosure_marshal_VOID__VOID,
-                  G_TYPE_NONE, 0);
-
-  signals[GOT_ROOMS] =
-    g_signal_new ("got-rooms",
-                  G_OBJECT_CLASS_TYPE (gabble_roomlist_channel_class),
-                  G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
-                  0,
-                  NULL, NULL,
-                  g_cclosure_marshal_VOID__BOXED,
-                  G_TYPE_NONE, 1, (dbus_g_type_get_collection ("GPtrArray", (dbus_g_type_get_struct ("GValueArray", G_TYPE_UINT, G_TYPE_STRING, (dbus_g_type_get_map ("GHashTable", G_TYPE_STRING, G_TYPE_VALUE)), G_TYPE_INVALID)))));
-
-  signals[LISTING_ROOMS] =
-    g_signal_new ("listing-rooms",
-                  G_OBJECT_CLASS_TYPE (gabble_roomlist_channel_class),
-                  G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
-                  0,
-                  NULL, NULL,
-                  g_cclosure_marshal_VOID__BOOLEAN,
-                  G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
-
-  dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (gabble_roomlist_channel_class), &dbus_glib_gabble_roomlist_channel_object_info);
 }
 
-static void
-room_list_flush_disco_pipeline_one (gpointer item, gpointer data)
-{
-  GabbleDiscoRequest *request = (GabbleDiscoRequest *) item;
-  GabbleDisco *disco = GABBLE_DISCO (data);
-
-  gabble_disco_cancel_request (disco, request);
-}
-
-void
-room_list_flush_disco_pipeline (GabbleRoomlistChannel *self)
-{
-  GabbleRoomlistChannelPrivate *priv = GABBLE_ROOMLIST_CHANNEL_GET_PRIVATE (self);
-
-  g_ptr_array_foreach (priv->disco_pipeline, (GFunc)
-      room_list_flush_disco_pipeline_one, priv->conn->disco);
-}
+static void stop_listing (GabbleRoomlistChannel *self);
 
 void
 gabble_roomlist_channel_dispose (GObject *object)
@@ -314,19 +274,18 @@ gabble_roomlist_channel_dispose (GObject *object)
 
   priv->dispose_has_run = TRUE;
 
-  if (priv->listing)
-    {
-      g_signal_emit (object, signals [LISTING_ROOMS], 0, FALSE);
-      priv->listing = FALSE;
-    }
+  stop_listing (self);
 
   if (!priv->closed)
     {
-      g_signal_emit (object, signals[CLOSED], 0);
+      tp_svc_channel_emit_closed ((TpSvcChannel *)object);
       priv->closed = TRUE;
     }
 
-  room_list_flush_disco_pipeline (self);
+  g_assert (priv->pending_room_signals != NULL);
+  g_assert (priv->pending_room_signals->len == 0);
+  g_ptr_array_free (priv->pending_room_signals, TRUE);
+  priv->pending_room_signals = NULL;
 
   if (G_OBJECT_CLASS (gabble_roomlist_channel_parent_class)->dispose)
     G_OBJECT_CLASS (gabble_roomlist_channel_parent_class)->dispose (object);
@@ -343,11 +302,8 @@ gabble_roomlist_channel_finalize (GObject *object)
   g_free (priv->object_path);
   g_free (priv->conference_server);
 
-  g_ptr_array_free (priv->disco_pipeline, TRUE);
-  g_hash_table_destroy (priv->remaining_rooms);
-
-  if (priv->signalled_rooms)
-    handle_set_destroy (priv->signalled_rooms);
+  if (priv->signalled_rooms != NULL)
+    tp_handle_set_destroy (priv->signalled_rooms);
 
   G_OBJECT_CLASS (gabble_roomlist_channel_parent_class)->finalize (object);
 }
@@ -368,93 +324,47 @@ _gabble_roomlist_channel_new (GabbleConnection *conn,
                     "conference-server", conference_server, NULL));
 }
 
-static void room_info_cb (GabbleDisco *, GabbleDiscoRequest *, const gchar *,
-    const gchar *, LmMessageNode *, GError *, gpointer);
-
 static gboolean
-return_true (gpointer key, gpointer value, gpointer data)
+emit_room_signal (gpointer data)
 {
+  GabbleRoomlistChannel *chan = data;
+  GabbleRoomlistChannelPrivate *priv = GABBLE_ROOMLIST_CHANNEL_GET_PRIVATE (chan);
+
+  if (!priv->listing)
+      return FALSE;
+
+  if (priv->pending_room_signals->len == 0)
+      return TRUE;
+
+  tp_svc_channel_type_room_list_emit_got_rooms (
+      (TpSvcChannelTypeRoomList *)chan, priv->pending_room_signals);
+
+  while (priv->pending_room_signals->len != 0)
+    {
+      gpointer boxed = g_ptr_array_index (priv->pending_room_signals, 0);
+      g_boxed_free (TP_TYPE_ROOM_STRUCT, boxed);
+      g_ptr_array_remove_index_fast (priv->pending_room_signals, 0);
+    }
+
   return TRUE;
 }
 
 static void
-room_list_fill_disco_pipeline (GabbleRoomlistChannel *chan)
-{
-  GabbleRoomlistChannelPrivate *priv =
-    GABBLE_ROOMLIST_CHANNEL_GET_PRIVATE (chan);
-
-  if (priv->closed)
-    {
-      DEBUG ("not refilling pipeline, channel is closed");
-    }
-  else
-    {
-      /* send disco requests for the JIDs in the remaining_rooms hash table
-       * until there are DISCO_PIPELINE_SIZE requests in progress */
-      while (priv->disco_pipeline->len < DISCO_PIPELINE_SIZE)
-        {
-          gchar *jid;
-          GabbleDiscoRequest *request;
-
-          jid = (gchar *) g_hash_table_find (priv->remaining_rooms,
-              return_true, NULL);
-          if (NULL == jid)
-            break;
-
-          request = gabble_disco_request (priv->conn->disco,
-              GABBLE_DISCO_TYPE_INFO, jid, NULL, room_info_cb, chan,
-              G_OBJECT(chan), NULL);
-
-          g_ptr_array_add (priv->disco_pipeline, request);
-
-          /* frees jid */
-          g_hash_table_remove (priv->remaining_rooms, jid);
-        }
-
-      if (0 == priv->disco_pipeline->len)
-        {
-          priv->listing = FALSE;
-          g_signal_emit (chan, signals [LISTING_ROOMS], 0, FALSE);
-        }
-    }
-}
-
-/**
- * destroy_value:
- * @data: a GValue to destroy
- *
- * destroys a GValue allocated on the heap
- */
-static void
-destroy_value (GValue *value)
-{
-  g_value_unset (value);
-  g_free (value);
-}
-
-static void
-room_info_cb (GabbleDisco *disco,
-              GabbleDiscoRequest *request,
-              const gchar *jid,
-              const gchar *node,
-              LmMessageNode *result,
-              GError *error,
-              gpointer user_data)
+room_info_cb (gpointer pipeline, GabbleDiscoItem *item, gpointer user_data)
 {
   GabbleRoomlistChannel *chan = user_data;
   GabbleRoomlistChannelPrivate *priv;
-  LmMessageNode *identity, *feature, *field, *value_node;
-  const char *category, *type, *var, *name, *value;
-  GabbleHandle handle;
+  TpBaseConnection *conn;
+  const char *jid, *category, *type, *var, *name;
+  TpHandle handle;
   GHashTable *keys;
   GValue room = {0,};
-  GPtrArray *rooms ;
   GValue *tmp;
-  gboolean is_muc;
+  gpointer k, v;
 
   #define INSERT_KEY(hash, name, type, type2, value) \
     do {\
-      tmp = g_new0 (GValue, 1); \
+      tmp = g_slice_new0 (GValue); \
       g_value_init (tmp, (type)); \
       g_value_set_##type2 (tmp, (value)); \
       g_hash_table_insert (hash, (name), tmp); \
@@ -462,209 +372,141 @@ room_info_cb (GabbleDisco *disco,
 
   g_assert (GABBLE_IS_ROOMLIST_CHANNEL (chan));
   priv = GABBLE_ROOMLIST_CHANNEL_GET_PRIVATE (chan);
+  conn = (TpBaseConnection *)priv->conn;
 
-  g_ptr_array_remove_fast (priv->disco_pipeline, request);
-
-  if (error)
-    {
-      DEBUG ("got error %s", error->message);
-      goto done;
-    }
-
-  identity = lm_message_node_get_child (result, "identity");
-  if (NULL == identity)
-    goto done;
-
-  name = lm_message_node_get_attribute (identity, "name");
-  if (NULL == name)
-    goto done;
-
-  category = lm_message_node_get_attribute (identity, "category");
-  if (NULL == category)
-    goto done;
-
-  type = lm_message_node_get_attribute (identity, "type");
-  if (NULL == type)
-    goto done;
+  jid = item->jid;
+  name = item->name;
+  category = item->category;
+  type = item->type;
 
   if (0 != strcmp (category, "conference") ||
       0 != strcmp (type, "text"))
-    goto done;
+    return;
+
+  if (!g_hash_table_lookup_extended (item->features, "http://jabber.org/protocol/muc", &k, &v))
+    {
+      /* not muc */
+      return;
+    }
 
   DEBUG ("got room identity, name=%s, category=%s, type=%s", name, category, type);
 
   keys = g_hash_table_new_full (g_str_hash, g_str_equal, NULL,
-                                (GDestroyNotify) destroy_value);
+                                (GDestroyNotify) tp_g_value_slice_free);
 
   INSERT_KEY (keys, "name", G_TYPE_STRING, string, name);
 
-  is_muc = FALSE;
+  if (g_hash_table_lookup_extended (item->features, "muc_membersonly", &k, &v))
+    INSERT_KEY (keys, "invite-only", G_TYPE_BOOLEAN, boolean, TRUE);
+  if (g_hash_table_lookup_extended (item->features, "muc_open", &k, &v))
+    INSERT_KEY (keys, "invite-only", G_TYPE_BOOLEAN, boolean, FALSE);
+  if (g_hash_table_lookup_extended (item->features, "muc_passwordprotected", &k, &v))
+    INSERT_KEY (keys, "password", G_TYPE_BOOLEAN, boolean, TRUE);
+  if (g_hash_table_lookup_extended (item->features, "muc_unsecure", &k, &v))
+    INSERT_KEY (keys, "password", G_TYPE_BOOLEAN, boolean, FALSE);
+  if (g_hash_table_lookup_extended (item->features, "muc_unsecured", &k, &v))
+    INSERT_KEY (keys, "password", G_TYPE_BOOLEAN, boolean, FALSE);
+  if (g_hash_table_lookup_extended (item->features, "muc_hidden", &k, &v))
+    INSERT_KEY (keys, "hidden", G_TYPE_BOOLEAN, boolean, TRUE);
+  if (g_hash_table_lookup_extended (item->features, "muc_public", &k, &v))
+    INSERT_KEY (keys, "hidden", G_TYPE_BOOLEAN, boolean, FALSE);
+  if (g_hash_table_lookup_extended (item->features, "muc_membersonly", &k, &v))
+    INSERT_KEY (keys, "members-only", G_TYPE_BOOLEAN, boolean, TRUE);
+  if (g_hash_table_lookup_extended (item->features, "muc_open", &k, &v))
+    INSERT_KEY (keys, "members-only", G_TYPE_BOOLEAN, boolean, FALSE);
+  if (g_hash_table_lookup_extended (item->features, "muc_moderated", &k, &v))
+    INSERT_KEY (keys, "moderated", G_TYPE_BOOLEAN, boolean, TRUE);
+  if (g_hash_table_lookup_extended (item->features, "muc_unmoderated", &k, &v))
+    INSERT_KEY (keys, "moderated", G_TYPE_BOOLEAN, boolean, FALSE);
+  if (g_hash_table_lookup_extended (item->features, "muc_nonanonymous", &k, &v))
+    INSERT_KEY (keys, "anonymous", G_TYPE_BOOLEAN, boolean, FALSE);
+  if (g_hash_table_lookup_extended (item->features, "muc_anonymous", &k, &v))
+    INSERT_KEY (keys, "anonymous", G_TYPE_BOOLEAN, boolean, TRUE);
+  if (g_hash_table_lookup_extended (item->features, "muc_semianonymous", &k, &v))
+    INSERT_KEY (keys, "anonymous", G_TYPE_BOOLEAN, boolean, FALSE);
+  if (g_hash_table_lookup_extended (item->features, "muc_persistent", &k, &v))
+    INSERT_KEY (keys, "persistent", G_TYPE_BOOLEAN, boolean, TRUE);
+  if (g_hash_table_lookup_extended (item->features, "muc_temporary", &k, &v))
+    INSERT_KEY (keys, "persistent", G_TYPE_BOOLEAN, boolean, FALSE);
 
-  for (feature = result->children; feature; feature = feature->next)
-    {
-      if (0 == strcmp (feature->name, "feature"))
-        {
-          var = lm_message_node_get_attribute (feature, "var");
-          if (!var)
-            continue;
-          if (0 == strcmp (var, "http://jabber.org/protocol/muc"))
-            is_muc = TRUE;
-          else if (0 == strcmp (var, "muc_membersonly"))
-            INSERT_KEY (keys, "invite-only", G_TYPE_BOOLEAN, boolean, TRUE);
-          else if (0 == strcmp (var, "muc_open"))
-            INSERT_KEY (keys, "invite-only", G_TYPE_BOOLEAN, boolean, FALSE);
-          else if (0 == strcmp (var, "muc_passwordprotected"))
-            INSERT_KEY (keys, "password", G_TYPE_BOOLEAN, boolean, TRUE);
-          else if (0 == strcmp (var, "muc_unsecure"))
-            INSERT_KEY (keys, "password", G_TYPE_BOOLEAN, boolean, FALSE);
-          else if (0 == strcmp (var, "muc_unsecured"))
-            INSERT_KEY (keys, "password", G_TYPE_BOOLEAN, boolean, FALSE);
-          else if (0 == strcmp (var, "muc_hidden"))
-            INSERT_KEY (keys, "hidden", G_TYPE_BOOLEAN, boolean, TRUE);
-          else if (0 == strcmp (var, "muc_public"))
-            INSERT_KEY (keys, "hidden", G_TYPE_BOOLEAN, boolean, FALSE);
-          else if (0 == strcmp (var, "muc_membersonly"))
-            INSERT_KEY (keys, "members-only", G_TYPE_BOOLEAN, boolean, TRUE);
-          else if (0 == strcmp (var, "muc_open"))
-            INSERT_KEY (keys, "members-only", G_TYPE_BOOLEAN, boolean, FALSE);
-          else if (0 == strcmp (var, "muc_moderated"))
-            INSERT_KEY (keys, "moderated", G_TYPE_BOOLEAN, boolean, TRUE);
-          else if (0 == strcmp (var, "muc_unmoderated"))
-            INSERT_KEY (keys, "moderated", G_TYPE_BOOLEAN, boolean, FALSE);
-          else if (0 == strcmp (var, "muc_nonanonymous"))
-            INSERT_KEY (keys, "anonymous", G_TYPE_BOOLEAN, boolean, FALSE);
-          else if (0 == strcmp (var, "muc_anonymous"))
-            INSERT_KEY (keys, "anonymous", G_TYPE_BOOLEAN, boolean, TRUE);
-          else if (0 == strcmp (var, "muc_semianonymous"))
-            INSERT_KEY (keys, "anonymous", G_TYPE_BOOLEAN, boolean, FALSE);
-          else if (0 == strcmp (var, "muc_persistent"))
-            INSERT_KEY (keys, "persistent", G_TYPE_BOOLEAN, boolean, TRUE);
-          else if (0 == strcmp (var, "muc_temporary"))
-            INSERT_KEY (keys, "persistent", G_TYPE_BOOLEAN, boolean, FALSE);
-          else
-            NODE_DEBUG (feature, "got unknown feature");
-        }
-      else if (0 == strcmp (feature->name, "x"))
-        {
-          if (lm_message_node_has_namespace (feature, NS_X_DATA, NULL))
-            {
-              for (field = feature->children;
-                   field; field = field->next)
-                {
-                  if (0 != strcmp (field->name, "field"))
-                    continue;
+  var = g_hash_table_lookup (item->features, "muc#roominfo_description");
+  if (var != NULL)
+    INSERT_KEY (keys, "description", G_TYPE_STRING, string, var);
 
-                  var = lm_message_node_get_attribute (field, "var");
-                  if (NULL == var)
-                    continue;
+  var = g_hash_table_lookup (item->features, "muc#roominfo_occupants");
+  if (var != NULL)
+    INSERT_KEY (keys, "members", G_TYPE_UINT, uint,
+                (guint) g_ascii_strtoull (var, NULL, 10));
 
-                  value_node = lm_message_node_get_child (field, "value");
-                  if (NULL == value_node)
-                    continue;
+  var = g_hash_table_lookup (item->features, "muc#roominfo_lang");
+  if (var != NULL)
+    INSERT_KEY (keys, "language", G_TYPE_STRING, string, var);
 
-                  value = lm_message_node_get_value (value_node);
-                  if (NULL == value)
-                    continue;
+  handle = gabble_handle_for_room (
+      conn->handles[TP_HANDLE_TYPE_ROOM], jid);
 
-                  if (0 == strcmp (var, "muc#roominfo_description"))
-                    {
-                      INSERT_KEY (keys, "description",
-                                  G_TYPE_STRING, string, value);
-                    }
-                  else if (0 == strcmp (var, "muc#roominfo_occupants"))
-                    {
-                      INSERT_KEY (keys, "members", G_TYPE_UINT, uint,
-                        (guint) g_ascii_strtoull (value, NULL, 10));
-                    }
-                  else if (0 == strcmp (var, "muc#roominfo_lang"))
-                    {
-                      INSERT_KEY (keys, "language", G_TYPE_STRING,
-                                  string, value);
-                    }
-                }
-            }
-        }
-    }
+  tp_handle_set_add (priv->signalled_rooms, handle);
 
-  if (is_muc)
-    {
-      DEBUG ("emitting new room signal for %s", jid);
+  g_value_init (&room, TP_TYPE_ROOM_STRUCT);
+  g_value_take_boxed (&room,
+      dbus_g_type_specialized_construct (TP_TYPE_ROOM_STRUCT));
 
-      handle = gabble_handle_for_room (priv->conn->handles, jid);
-
-      handle_set_add (priv->signalled_rooms, handle);
-
-      g_value_init (&room, TP_TYPE_ROOM_STRUCT);
-      g_value_take_boxed (&room,
-          dbus_g_type_specialized_construct (TP_TYPE_ROOM_STRUCT));
-
-      dbus_g_type_struct_set (&room,
-          0, handle,
-          1, "org.freedesktop.Telepathy.Channel.Type.Text",
-          2, keys,
-          G_MAXUINT);
-
-      rooms = g_ptr_array_sized_new (1);
-      g_ptr_array_add (rooms, g_value_get_boxed (&room));
-      g_signal_emit (chan, signals[GOT_ROOMS], 0, rooms);
-      g_ptr_array_free (rooms, TRUE);
-      g_value_unset (&room);
-    }
+  dbus_g_type_struct_set (&room,
+      0, handle,
+      1, "org.freedesktop.Telepathy.Channel.Type.Text",
+      2, keys,
+      G_MAXUINT);
 
+  DEBUG ("adding new room signal data to pending: %s", jid);
+  g_ptr_array_add (priv->pending_room_signals, g_value_get_boxed (&room));
   g_hash_table_destroy (keys);
-
-done:
-  room_list_fill_disco_pipeline (chan);
-
-  return;
 }
 
 static void
-rooms_cb (GabbleDisco *disco,
-          GabbleDiscoRequest *request,
-          const gchar *jid,
-          const gchar *node,
-          LmMessageNode *result,
-          GError *error,
-          gpointer user_data)
+rooms_end_cb (gpointer data, gpointer user_data)
 {
-  LmMessageNode *iter;
   GabbleRoomlistChannel *chan = user_data;
-  GabbleRoomlistChannelPrivate *priv;
-  const char *item_jid;
-  gpointer key, value;
+  GabbleRoomlistChannelPrivate *priv = GABBLE_ROOMLIST_CHANNEL_GET_PRIVATE (chan);
 
-  g_assert (GABBLE_IS_ROOMLIST_CHANNEL (chan));
+  emit_room_signal (chan);
 
-  priv = GABBLE_ROOMLIST_CHANNEL_GET_PRIVATE (chan);
+  priv->listing = FALSE;
+  tp_svc_channel_type_room_list_emit_listing_rooms (
+      (TpSvcChannelTypeRoomList *)chan, FALSE);
 
-  if (error)
-    {
-      DEBUG ("got error %s", error->message);
-      goto out;
-    }
-
-  iter = result->children;
-
-  for (; iter; iter = iter->next)
-    {
-      if (0 != strcmp (iter->name, "item"))
-        continue;
-
-      item_jid = lm_message_node_get_attribute (iter, "jid");
-
-      if (NULL != item_jid &&
-          !g_hash_table_lookup_extended (priv->remaining_rooms, item_jid, &key, &value))
-        {
-          gchar *tmp = g_strdup (item_jid);
-          g_hash_table_insert (priv->remaining_rooms, tmp, tmp);
-        }
-    }
-
-out:
-  room_list_fill_disco_pipeline (chan);
+  g_source_remove (priv->timer_source_id);
+  priv->timer_source_id = 0;
 }
 
+static void
+stop_listing (GabbleRoomlistChannel *self)
+{
+  GabbleRoomlistChannelPrivate *priv = GABBLE_ROOMLIST_CHANNEL_GET_PRIVATE (self);
+
+  if (priv->listing)
+    {
+      emit_room_signal (self);
+      priv->listing = FALSE;
+      tp_svc_channel_type_room_list_emit_listing_rooms (
+          (TpSvcChannelTypeRoomList *)self, FALSE);
+    }
+
+  if (priv->disco_pipeline != NULL)
+    {
+      gabble_disco_pipeline_destroy (priv->disco_pipeline);
+      priv->disco_pipeline = NULL;
+    }
+
+  if (priv->timer_source_id)
+    {
+      g_source_remove (priv->timer_source_id);
+      priv->timer_source_id = 0;
+    }
+
+  g_assert (priv->pending_room_signals->len == 0);
+}
+
+
 /************************* D-Bus Method definitions **************************/
 
 /**
@@ -672,24 +514,19 @@ out:
  *
  * Implements D-Bus method Close
  * on interface org.freedesktop.Telepathy.Channel
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
  */
-gboolean
-gabble_roomlist_channel_close (GabbleRoomlistChannel *self,
-                               GError **error)
+static void
+gabble_roomlist_channel_close (TpSvcChannel *iface,
+                               DBusGMethodInvocation *context)
 {
+  GabbleRoomlistChannel *self = GABBLE_ROOMLIST_CHANNEL (iface);
   g_assert (GABBLE_IS_ROOMLIST_CHANNEL (self));
 
   DEBUG ("called on %p", self);
 
   g_object_run_dispose (G_OBJECT (self));
 
-  return TRUE;
+  tp_svc_channel_return_from_close (context);
 }
 
 
@@ -698,20 +535,13 @@ gabble_roomlist_channel_close (GabbleRoomlistChannel *self,
  *
  * Implements D-Bus method GetChannelType
  * on interface org.freedesktop.Telepathy.Channel
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
  */
-gboolean
-gabble_roomlist_channel_get_channel_type (GabbleRoomlistChannel *self,
-                                          gchar **ret,
-                                          GError **error)
+static void
+gabble_roomlist_channel_get_channel_type (TpSvcChannel *iface,
+                                          DBusGMethodInvocation *context)
 {
-  *ret = g_strdup (TP_IFACE_CHANNEL_TYPE_ROOM_LIST);
-  return TRUE;
+  tp_svc_channel_return_from_get_channel_type (context,
+      TP_IFACE_CHANNEL_TYPE_ROOM_LIST);
 }
 
 
@@ -720,25 +550,12 @@ gabble_roomlist_channel_get_channel_type (GabbleRoomlistChannel *self,
  *
  * Implements D-Bus method GetHandle
  * on interface org.freedesktop.Telepathy.Channel
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
  */
-gboolean
-gabble_roomlist_channel_get_handle (GabbleRoomlistChannel *self,
-                                    guint *ret,
-                                    guint *ret1,
-                                    GError **error)
+static void
+gabble_roomlist_channel_get_handle (TpSvcChannel *self,
+                                    DBusGMethodInvocation *context)
 {
-  g_assert (GABBLE_IS_ROOMLIST_CHANNEL (self));
-
-  *ret = 0;
-  *ret1 = 0;
-
-  return TRUE;
+  tp_svc_channel_return_from_get_handle (context, 0, 0);
 }
 
 
@@ -747,23 +564,14 @@ gabble_roomlist_channel_get_handle (GabbleRoomlistChannel *self,
  *
  * Implements D-Bus method GetInterfaces
  * on interface org.freedesktop.Telepathy.Channel
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
  */
-gboolean
-gabble_roomlist_channel_get_interfaces (GabbleRoomlistChannel *self,
-                                        gchar ***ret,
-                                        GError **error)
+static void
+gabble_roomlist_channel_get_interfaces (TpSvcChannel *iface,
+                                        DBusGMethodInvocation *context)
 {
   const char *interfaces[] = { NULL };
 
-  *ret = g_strdupv ((gchar **) interfaces);
-
-  return TRUE;
+  tp_svc_channel_return_from_get_interfaces (context, interfaces);
 }
 
 
@@ -779,18 +587,18 @@ gabble_roomlist_channel_get_interfaces (GabbleRoomlistChannel *self,
  *
  * Returns: TRUE if successful, FALSE if an error was thrown.
  */
-gboolean
-gabble_roomlist_channel_get_listing_rooms (GabbleRoomlistChannel *self,
-                                           gboolean *ret,
-                                           GError **error)
+static void
+gabble_roomlist_channel_get_listing_rooms (TpSvcChannelTypeRoomList *iface,
+                                           DBusGMethodInvocation *context)
 {
+  GabbleRoomlistChannel *self = GABBLE_ROOMLIST_CHANNEL (iface);
   GabbleRoomlistChannelPrivate *priv;
 
   g_assert (GABBLE_IS_ROOMLIST_CHANNEL (self));
 
   priv = GABBLE_ROOMLIST_CHANNEL_GET_PRIVATE (self);
-  *ret = priv->listing;
-  return TRUE;
+  tp_svc_channel_type_room_list_return_from_get_listing_rooms (
+      context, priv->listing);
 }
 
 
@@ -806,10 +614,11 @@ gabble_roomlist_channel_get_listing_rooms (GabbleRoomlistChannel *self,
  *
  * Returns: TRUE if successful, FALSE if an error was thrown.
  */
-gboolean
-gabble_roomlist_channel_list_rooms (GabbleRoomlistChannel *self,
-                                    GError **error)
+static void
+gabble_roomlist_channel_list_rooms (TpSvcChannelTypeRoomList *iface,
+                                    DBusGMethodInvocation *context)
 {
+  GabbleRoomlistChannel *self = GABBLE_ROOMLIST_CHANNEL (iface);
   GabbleRoomlistChannelPrivate *priv;
 
   g_assert (GABBLE_IS_ROOMLIST_CHANNEL (self));
@@ -817,10 +626,60 @@ gabble_roomlist_channel_list_rooms (GabbleRoomlistChannel *self,
   priv = GABBLE_ROOMLIST_CHANNEL_GET_PRIVATE (self);
 
   priv->listing = TRUE;
-  g_signal_emit (self, signals[LISTING_ROOMS], 0, TRUE);
-  gabble_disco_request (priv->conn->disco, GABBLE_DISCO_TYPE_ITEMS,
-                        priv->conference_server, NULL,
-                        rooms_cb, self, G_OBJECT(self), NULL);
-  return TRUE;
+  tp_svc_channel_type_room_list_emit_listing_rooms (iface, TRUE);
+
+  if (priv->disco_pipeline == NULL)
+    priv->disco_pipeline = gabble_disco_pipeline_init (priv->conn->disco,
+        room_info_cb, rooms_end_cb, self);
+
+  gabble_disco_pipeline_run (priv->disco_pipeline, priv->conference_server);
+
+  priv->timer_source_id = g_timeout_add (ROOM_SIGNAL_INTERVAL,
+      emit_room_signal, self);
+
+  tp_svc_channel_type_room_list_return_from_list_rooms (context);
 }
 
+/**
+ * gabble_roomlist_channel_stop_listing
+ *
+ * Implements D-Bus method StopListing
+ * on interface org.freedesktop.Telepathy.Channel.Type.RoomList
+ */
+static void
+gabble_roomlist_channel_stop_listing (TpSvcChannelTypeRoomList *iface,
+                                      DBusGMethodInvocation *context)
+{
+  GabbleRoomlistChannel *self = GABBLE_ROOMLIST_CHANNEL (iface);
+
+  g_assert (GABBLE_IS_ROOMLIST_CHANNEL (self));
+
+  stop_listing (self);
+}
+
+static void
+channel_iface_init(gpointer g_iface, gpointer iface_data)
+{
+  TpSvcChannelClass *klass = (TpSvcChannelClass *)g_iface;
+
+#define IMPLEMENT(x) tp_svc_channel_implement_##x (\
+    klass, gabble_roomlist_channel_##x)
+  IMPLEMENT(close);
+  IMPLEMENT(get_channel_type);
+  IMPLEMENT(get_handle);
+  IMPLEMENT(get_interfaces);
+#undef IMPLEMENT
+}
+
+static void
+roomlist_iface_init(gpointer g_iface, gpointer iface_data)
+{
+  TpSvcChannelTypeRoomListClass *klass = (TpSvcChannelTypeRoomListClass *)g_iface;
+
+#define IMPLEMENT(x) tp_svc_channel_type_room_list_implement_##x (\
+    klass, gabble_roomlist_channel_##x)
+  IMPLEMENT(get_listing_rooms);
+  IMPLEMENT(list_rooms);
+  IMPLEMENT(stop_listing);
+#undef IMPLEMENT
+}
