@@ -1,6 +1,6 @@
 /*
  * gabble-roster-channel.c - Source for GabbleRosterChannel
- * Copyright (C) 2005 Collabora Ltd.
+ * Copyright (C) 2005, 2006 Collabora Ltd.
  * Copyright (C) 2005 Nokia Corporation
  *
  * This library is free software; you can redistribute it and/or
@@ -26,33 +26,27 @@
 
 #include "debug.h"
 #include "gabble-connection.h"
-#include "gintset.h"
-#include "group-mixin.h"
-#include "handle-set.h"
+#include <telepathy-glib/group-mixin.h>
+#include "handles.h"
 #include "roster.h"
-#include "telepathy-errors.h"
-#include "telepathy-helpers.h"
-#include "telepathy-interfaces.h"
-#include "tp-channel-iface.h"
+#include <telepathy-glib/dbus.h>
+#include <telepathy-glib/errors.h>
+#include <telepathy-glib/interfaces.h>
+#include <telepathy-glib/channel-iface.h>
+#include <telepathy-glib/svc-channel.h>
 #include "util.h"
 
 #include "gabble-roster-channel.h"
-#include "gabble-roster-channel-glue.h"
-#include "gabble-roster-channel-signals-marshal.h"
+
+static void channel_iface_init (gpointer, gpointer);
 
 G_DEFINE_TYPE_WITH_CODE (GabbleRosterChannel, gabble_roster_channel,
-    G_TYPE_OBJECT, G_IMPLEMENT_INTERFACE (TP_TYPE_CHANNEL_IFACE, NULL));
-
-/* signal enum */
-enum
-{
-    CLOSED,
-    GROUP_FLAGS_CHANGED,
-    MEMBERS_CHANGED,
-    LAST_SIGNAL
-};
-
-static guint signals[LAST_SIGNAL] = {0};
+    G_TYPE_OBJECT,
+    G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_CHANNEL, channel_iface_init);
+    G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_CHANNEL_INTERFACE_GROUP,
+      tp_group_mixin_iface_init);
+    G_IMPLEMENT_INTERFACE (TP_TYPE_SVC_CHANNEL_TYPE_CONTACT_LIST, NULL);
+    G_IMPLEMENT_INTERFACE (TP_TYPE_CHANNEL_IFACE, NULL));
 
 /* properties */
 enum
@@ -72,9 +66,11 @@ struct _GabbleRosterChannelPrivate
 {
   GabbleConnection *conn;
   char *object_path;
-  GabbleHandle handle;
+  TpHandle handle;
+  guint handle_type;
 
   gboolean dispose_has_run;
+  gboolean closed;
 };
 
 #define GABBLE_ROSTER_CHANNEL_GET_PRIVATE(obj) \
@@ -98,31 +94,50 @@ gabble_roster_channel_constructor (GType type, guint n_props,
   GObject *obj;
   GabbleRosterChannelPrivate *priv;
   DBusGConnection *bus;
-  GabbleHandleRepo *handles;
+  TpBaseConnection *conn;
   gboolean valid;
-  GabbleHandle self_handle;
+  TpHandle self_handle;
+  guint handle_type;
 
   obj = G_OBJECT_CLASS (gabble_roster_channel_parent_class)->
            constructor (type, n_props, props);
   priv = GABBLE_ROSTER_CHANNEL_GET_PRIVATE (GABBLE_ROSTER_CHANNEL (obj));
-  handles = priv->conn->handles;
-  self_handle = priv->conn->self_handle;
+  conn = (TpBaseConnection *)priv->conn;
+  handle_type = priv->handle_type;
+  self_handle = conn->self_handle;
 
   /* register object on the bus */
   bus = tp_get_bus ();
   dbus_g_connection_register_g_object (bus, priv->object_path, obj);
 
+  g_assert (handle_type == TP_HANDLE_TYPE_GROUP
+      || handle_type == TP_HANDLE_TYPE_LIST);
+
   /* ref our list handle */
-  valid = gabble_handle_ref (handles, TP_HANDLE_TYPE_LIST, priv->handle);
+  valid = tp_handle_ref (conn->handles[handle_type], priv->handle);
   g_assert (valid);
 
   /* initialize group mixin */
-  gabble_group_mixin_init (obj, G_STRUCT_OFFSET (GabbleRosterChannel, group),
-                           handles, self_handle);
+  tp_group_mixin_init ((TpSvcChannelInterfaceGroup *)obj,
+      G_STRUCT_OFFSET (GabbleRosterChannel, group),
+      conn->handles[TP_HANDLE_TYPE_CONTACT],
+      self_handle);
 
-  if (GABBLE_LIST_HANDLE_PUBLISH == priv->handle)
+  if (handle_type == TP_HANDLE_TYPE_GROUP)
     {
-      gabble_group_mixin_change_flags (obj,
+      tp_group_mixin_change_flags ((TpSvcChannelInterfaceGroup *)obj,
+          TP_CHANNEL_GROUP_FLAG_CAN_ADD |
+          TP_CHANNEL_GROUP_FLAG_CAN_REMOVE,
+          0);
+    }
+  else if (handle_type != TP_HANDLE_TYPE_LIST)
+    {
+      g_assert_not_reached ();
+    }
+  /* magic contact lists from here down... */
+  else if (GABBLE_LIST_HANDLE_PUBLISH == priv->handle)
+    {
+      tp_group_mixin_change_flags ((TpSvcChannelInterfaceGroup *)obj,
           TP_CHANNEL_GROUP_FLAG_CAN_REMOVE |
           TP_CHANNEL_GROUP_FLAG_MESSAGE_ACCEPT |
           TP_CHANNEL_GROUP_FLAG_MESSAGE_REMOVE,
@@ -130,7 +145,7 @@ gabble_roster_channel_constructor (GType type, guint n_props,
     }
   else if (GABBLE_LIST_HANDLE_SUBSCRIBE == priv->handle)
     {
-      gabble_group_mixin_change_flags (obj,
+      tp_group_mixin_change_flags ((TpSvcChannelInterfaceGroup *)obj,
           TP_CHANNEL_GROUP_FLAG_CAN_ADD |
           TP_CHANNEL_GROUP_FLAG_CAN_REMOVE |
           TP_CHANNEL_GROUP_FLAG_CAN_RESCIND |
@@ -141,13 +156,13 @@ gabble_roster_channel_constructor (GType type, guint n_props,
     }
   else if (GABBLE_LIST_HANDLE_KNOWN == priv->handle)
     {
-      gabble_group_mixin_change_flags (obj,
+      tp_group_mixin_change_flags ((TpSvcChannelInterfaceGroup *)obj,
           TP_CHANNEL_GROUP_FLAG_CAN_REMOVE,
           0);
     }
   else if (GABBLE_LIST_HANDLE_DENY == priv->handle)
     {
-      gabble_group_mixin_change_flags (obj,
+      tp_group_mixin_change_flags ((TpSvcChannelInterfaceGroup *)obj,
           TP_CHANNEL_GROUP_FLAG_CAN_ADD |
           TP_CHANNEL_GROUP_FLAG_CAN_REMOVE,
           0);
@@ -177,7 +192,7 @@ gabble_roster_channel_get_property (GObject    *object,
       g_value_set_static_string (value, TP_IFACE_CHANNEL_TYPE_CONTACT_LIST);
       break;
     case PROP_HANDLE_TYPE:
-      g_value_set_uint (value, TP_HANDLE_TYPE_LIST);
+      g_value_set_uint (value, priv->handle_type);
       break;
     case PROP_HANDLE:
       g_value_set_uint (value, priv->handle);
@@ -205,6 +220,9 @@ gabble_roster_channel_set_property (GObject     *object,
       g_free (priv->object_path);
       priv->object_path = g_value_dup_string (value);
       break;
+    case PROP_HANDLE_TYPE:
+      priv->handle_type = g_value_get_uint (value);
+      break;
     case PROP_HANDLE:
       priv->handle = g_value_get_uint (value);
       break;
@@ -220,8 +238,8 @@ gabble_roster_channel_set_property (GObject     *object,
 static void gabble_roster_channel_dispose (GObject *object);
 static void gabble_roster_channel_finalize (GObject *object);
 
-static gboolean _gabble_roster_channel_add_member_cb (GObject *obj, GabbleHandle handle, const gchar *message, GError **error);
-static gboolean _gabble_roster_channel_remove_member_cb (GObject *obj, GabbleHandle handle, const gchar *message, GError **error);
+static gboolean _gabble_roster_channel_add_member_cb (TpSvcChannelInterfaceGroup *obj, TpHandle handle, const gchar *message, GError **error);
+static gboolean _gabble_roster_channel_remove_member_cb (TpSvcChannelInterfaceGroup *obj, TpHandle handle, const gchar *message, GError **error);
 
 static void
 gabble_roster_channel_class_init (GabbleRosterChannelClass *gabble_roster_channel_class)
@@ -254,21 +272,10 @@ gabble_roster_channel_class_init (GabbleRosterChannelClass *gabble_roster_channe
   g_object_class_override_property (object_class, PROP_HANDLE_TYPE, "handle-type");
   g_object_class_override_property (object_class, PROP_HANDLE, "handle");
 
-  signals[CLOSED] =
-    g_signal_new ("closed",
-                  G_OBJECT_CLASS_TYPE (gabble_roster_channel_class),
-                  G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
-                  0,
-                  NULL, NULL,
-                  g_cclosure_marshal_VOID__VOID,
-                  G_TYPE_NONE, 0);
-
-  gabble_group_mixin_class_init (object_class,
-                                 G_STRUCT_OFFSET (GabbleRosterChannelClass, group_class),
-                                 _gabble_roster_channel_add_member_cb,
-                                 _gabble_roster_channel_remove_member_cb);
-
-  dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (gabble_roster_channel_class), &dbus_glib_gabble_roster_channel_object_info);
+  tp_group_mixin_class_init ((TpSvcChannelInterfaceGroupClass *)object_class,
+                             G_STRUCT_OFFSET (GabbleRosterChannelClass, group_class),
+                             _gabble_roster_channel_add_member_cb,
+                             _gabble_roster_channel_remove_member_cb);
 }
 
 void
@@ -282,7 +289,8 @@ gabble_roster_channel_dispose (GObject *object)
 
   priv->dispose_has_run = TRUE;
 
-  g_signal_emit(self, signals[CLOSED], 0);
+  if (!priv->closed)
+    tp_svc_channel_emit_closed ((TpSvcChannel *)object);
 
   /* release any references held by the object here */
 
@@ -295,14 +303,15 @@ gabble_roster_channel_finalize (GObject *object)
 {
   GabbleRosterChannel *self = GABBLE_ROSTER_CHANNEL (object);
   GabbleRosterChannelPrivate *priv = GABBLE_ROSTER_CHANNEL_GET_PRIVATE (self);
+  TpBaseConnection *conn = (TpBaseConnection *)priv->conn;
 
   /* free any data held directly by the object here */
 
   g_free (priv->object_path);
 
-  gabble_handle_unref (priv->conn->handles, TP_HANDLE_TYPE_LIST, priv->handle);
+  tp_handle_unref (conn->handles[priv->handle_type], priv->handle);
 
-  gabble_group_mixin_finalize (object);
+  tp_group_mixin_finalize ((TpSvcChannelInterfaceGroup *)object);
 
   G_OBJECT_CLASS (gabble_roster_channel_parent_class)->finalize (object);
 }
@@ -311,19 +320,21 @@ gabble_roster_channel_finalize (GObject *object)
 static gboolean
 _gabble_roster_channel_send_presence (GabbleRosterChannel *chan,
                                       LmMessageSubType sub_type,
-                                      GabbleHandle handle,
+                                      TpHandle handle,
                                       const gchar *status,
                                       GError **error)
 {
   GabbleRosterChannelPrivate *priv;
-  GabbleHandleRepo *repo;
+  TpBaseConnection *conn;
+  TpHandleRepoIface *repo;
   const char *contact;
   LmMessage *message;
   gboolean result;
 
   priv = GABBLE_ROSTER_CHANNEL_GET_PRIVATE (chan);
-  repo = priv->conn->handles;
-  contact = gabble_handle_inspect (repo, TP_HANDLE_TYPE_CONTACT, handle);
+  conn = (TpBaseConnection *)priv->conn;
+  repo = conn->handles[TP_HANDLE_TYPE_CONTACT];
+  contact = tp_handle_inspect (repo, handle);
 
   message = lm_message_new_with_sub_type (contact,
       LM_MESSAGE_TYPE_PRESENCE,
@@ -349,24 +360,37 @@ _gabble_roster_channel_send_presence (GabbleRosterChannel *chan,
  * Called by the group mixin to add one member.
  */
 static gboolean
-_gabble_roster_channel_add_member_cb (GObject *obj,
-                                      GabbleHandle handle,
+_gabble_roster_channel_add_member_cb (TpSvcChannelInterfaceGroup *obj,
+                                      TpHandle handle,
                                       const gchar *message,
                                       GError **error)
 {
   GabbleRosterChannelPrivate *priv;
-  GabbleHandleRepo *repo;
+  TpBaseConnection *conn;
   gboolean ret = FALSE;
 
   priv = GABBLE_ROSTER_CHANNEL_GET_PRIVATE (GABBLE_ROSTER_CHANNEL (obj));
+  conn = (TpBaseConnection *)priv->conn;
 
-  repo = priv->conn->handles;
-
-  DEBUG ("called on %s with handle %u (%s) \"%s\"", gabble_handle_inspect (repo, TP_HANDLE_TYPE_LIST, priv->handle), handle,
-      gabble_handle_inspect (repo, TP_HANDLE_TYPE_CONTACT, handle), message);
+  DEBUG ("called on %s with handle %u (%s) \"%s\"",
+      tp_handle_inspect (conn->handles[priv->handle_type],
+        priv->handle), handle,
+      tp_handle_inspect (conn->handles[TP_HANDLE_TYPE_CONTACT],
+        handle), message);
 
+  if (TP_HANDLE_TYPE_GROUP == priv->handle_type)
+    {
+      ret = gabble_roster_handle_add_to_group (priv->conn->roster,
+          handle, priv->handle, error);
+    }
+  else if (TP_HANDLE_TYPE_LIST != priv->handle_type)
+    {
+      g_assert_not_reached ();
+      return FALSE;
+    }
+  /* "magic" contact lists, from here down... */
   /* publish list */
-  if (GABBLE_LIST_HANDLE_PUBLISH == priv->handle)
+  else if (GABBLE_LIST_HANDLE_PUBLISH == priv->handle)
     {
       /* send <presence type="subscribed"> */
       ret = _gabble_roster_channel_send_presence (GABBLE_ROSTER_CHANNEL (obj),
@@ -375,6 +399,9 @@ _gabble_roster_channel_add_member_cb (GObject *obj,
   /* subscribe list */
   else if (GABBLE_LIST_HANDLE_SUBSCRIBE == priv->handle)
     {
+      /* add item to the roster (GTalk depends on this, clearing the H flag) */
+      gabble_roster_handle_add (priv->conn->roster, handle, NULL);
+
       /* send <presence type="subscribe"> */
       ret = _gabble_roster_channel_send_presence (GABBLE_ROSTER_CHANNEL (obj),
           LM_MESSAGE_SUB_TYPE_SUBSCRIBE, handle, message, error);
@@ -401,28 +428,54 @@ _gabble_roster_channel_add_member_cb (GObject *obj,
  * Called by the group mixin to remove one member.
  */
 static gboolean
-_gabble_roster_channel_remove_member_cb (GObject *obj,
-                                         GabbleHandle handle,
+_gabble_roster_channel_remove_member_cb (TpSvcChannelInterfaceGroup *obj,
+                                         TpHandle handle,
                                          const gchar *message,
                                          GError **error)
 {
   GabbleRosterChannelPrivate *priv;
-  GabbleHandleRepo *repo;
+  TpBaseConnection *conn;
   gboolean ret = FALSE;
 
   priv = GABBLE_ROSTER_CHANNEL_GET_PRIVATE (GABBLE_ROSTER_CHANNEL (obj));
+  conn = (TpBaseConnection *)priv->conn;
 
-  repo = priv->conn->handles;
-
-  DEBUG ("called on %s with handle %u (%s) \"%s\"", gabble_handle_inspect (repo, TP_HANDLE_TYPE_LIST, priv->handle), handle,
-      gabble_handle_inspect (repo, TP_HANDLE_TYPE_CONTACT, handle), message);
+  DEBUG ("called on %s with handle %u (%s) \"%s\"", tp_handle_inspect (
+         conn->handles[priv->handle_type], priv->handle), handle,
+      tp_handle_inspect (conn->handles[TP_HANDLE_TYPE_CONTACT],
+        handle), message);
 
+  if (TP_HANDLE_TYPE_GROUP == priv->handle_type)
+    {
+      ret = gabble_roster_handle_remove_from_group (priv->conn->roster,
+          handle, priv->handle, error);
+    }
+  else if (TP_HANDLE_TYPE_LIST != priv->handle_type)
+    {
+      g_assert_not_reached ();
+      return FALSE;
+    }
+  /* "magic" contact lists, from here down... */
   /* publish list */
-  if (GABBLE_LIST_HANDLE_PUBLISH == priv->handle)
+  else if (GABBLE_LIST_HANDLE_PUBLISH == priv->handle)
     {
       /* send <presence type="unsubscribed"> */
       ret = _gabble_roster_channel_send_presence (GABBLE_ROSTER_CHANNEL (obj),
           LM_MESSAGE_SUB_TYPE_UNSUBSCRIBED, handle, message, error);
+
+      /* remove it from local_pending here, because roster callback doesn't
+         know if it can (subscription='none' is used both during request and
+         when it's rejected) */
+      if (tp_handle_set_is_member (GABBLE_ROSTER_CHANNEL (obj)->group.local_pending, handle))
+        {
+          TpIntSet *rem = tp_intset_new ();
+
+          tp_intset_add (rem, handle);
+          tp_group_mixin_change_members (obj, "", NULL, rem, NULL, NULL,
+              0, 0);
+
+          tp_intset_destroy (rem);
+        }
     }
   /* subscribe list */
   else if (GABBLE_LIST_HANDLE_SUBSCRIBE == priv->handle)
@@ -453,73 +506,51 @@ _gabble_roster_channel_remove_member_cb (GObject *obj,
 }
 
 
-/**
- * gabble_roster_channel_add_members
- *
- * Implements D-Bus method AddMembers
- * on interface org.freedesktop.Telepathy.Channel.Interface.Group
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
- */
-gboolean
-gabble_roster_channel_add_members (GabbleRosterChannel *self,
-                                   const GArray *contacts,
-                                   const gchar *message,
-                                   GError **error)
-{
-  return gabble_group_mixin_add_members (G_OBJECT (self), contacts, message,
-      error);
-}
-
-
 /**
  * gabble_roster_channel_close
  *
  * Implements D-Bus method Close
  * on interface org.freedesktop.Telepathy.Channel
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
  */
-gboolean
-gabble_roster_channel_close (GabbleRosterChannel *self,
-                             GError **error)
+static void
+gabble_roster_channel_close (TpSvcChannel *iface,
+                             DBusGMethodInvocation *context)
 {
-  *error = g_error_new (TELEPATHY_ERRORS, NotImplemented,
-                        "you may not close contact list channels");
+  GabbleRosterChannel *self = GABBLE_ROSTER_CHANNEL (iface);
+  GabbleRosterChannelPrivate *priv;
 
-  return FALSE;
-}
+  g_assert (GABBLE_IS_ROSTER_CHANNEL (self));
 
+  priv = GABBLE_ROSTER_CHANNEL_GET_PRIVATE (self);
 
-/**
- * gabble_roster_channel_get_all_members
- *
- * Implements D-Bus method GetAllMembers
- * on interface org.freedesktop.Telepathy.Channel.Interface.Group
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
- */
-gboolean
-gabble_roster_channel_get_all_members (GabbleRosterChannel *self,
-                                       GArray **ret,
-                                       GArray **ret1,
-                                       GArray **ret2,
-                                       GError **error)
-{
-  return gabble_group_mixin_get_all_members (G_OBJECT (self), ret, ret1, ret2,
-      error);
+  if (priv->handle_type == TP_HANDLE_TYPE_LIST)
+    {
+      GError e = { TP_ERRORS, TP_ERROR_NOT_IMPLEMENTED,
+          "you may not close contact list channels" };
+
+      dbus_g_method_return_error (context, &e);
+    }
+  else /* TP_HANDLE_TYPE_GROUP */
+    {
+      if (tp_handle_set_size (self->group.members) == 0)
+        {
+          /* deleting groups isn't a concept that exists on XMPP,
+           * so just close the channel */
+
+          priv->closed = TRUE;
+          tp_svc_channel_emit_closed ((TpSvcChannel *)self);
+          tp_svc_channel_return_from_close (context);
+          return;
+        }
+
+      else
+        {
+          GError e = { TP_ERRORS, TP_ERROR_NOT_AVAILABLE,
+              "you may not close this group, because it's not empty" };
+
+          dbus_g_method_return_error (context, &e);
+        }
+    }
 }
 
 
@@ -528,42 +559,13 @@ gabble_roster_channel_get_all_members (GabbleRosterChannel *self,
  *
  * Implements D-Bus method GetChannelType
  * on interface org.freedesktop.Telepathy.Channel
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
  */
-gboolean
-gabble_roster_channel_get_channel_type (GabbleRosterChannel *self,
-                                        gchar **ret,
-                                        GError **error)
+static void
+gabble_roster_channel_get_channel_type (TpSvcChannel *iface,
+                                        DBusGMethodInvocation *context)
 {
-  *ret = g_strdup (TP_IFACE_CHANNEL_TYPE_CONTACT_LIST);
-
-  return TRUE;
-}
-
-
-/**
- * gabble_roster_channel_get_group_flags
- *
- * Implements D-Bus method GetGroupFlags
- * on interface org.freedesktop.Telepathy.Channel.Interface.Group
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
- */
-gboolean
-gabble_roster_channel_get_group_flags (GabbleRosterChannel *self,
-                                       guint *ret,
-                                       GError **error)
-{
-  return gabble_group_mixin_get_group_flags (G_OBJECT (self), ret, error);
+  tp_svc_channel_return_from_get_channel_type (context,
+      TP_IFACE_CHANNEL_TYPE_CONTACT_LIST);
 }
 
 
@@ -572,52 +574,20 @@ gabble_roster_channel_get_group_flags (GabbleRosterChannel *self,
  *
  * Implements D-Bus method GetHandle
  * on interface org.freedesktop.Telepathy.Channel
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
  */
-gboolean
-gabble_roster_channel_get_handle (GabbleRosterChannel *self,
-                                  guint *ret,
-                                  guint *ret1,
-                                  GError **error)
+static void
+gabble_roster_channel_get_handle (TpSvcChannel *iface,
+                                  DBusGMethodInvocation *context)
 {
+  GabbleRosterChannel *self = GABBLE_ROSTER_CHANNEL (iface);
   GabbleRosterChannelPrivate *priv;
 
   g_assert (GABBLE_IS_ROSTER_CHANNEL (self));
 
   priv = GABBLE_ROSTER_CHANNEL_GET_PRIVATE (self);
 
-  *ret = TP_HANDLE_TYPE_LIST;
-  *ret1 = priv->handle;
-
-  return TRUE;
-}
-
-
-/**
- * gabble_roster_channel_get_handle_owners
- *
- * Implements D-Bus method GetHandleOwners
- * on interface org.freedesktop.Telepathy.Channel.Interface.Group
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
- */
-gboolean
-gabble_roster_channel_get_handle_owners (GabbleRosterChannel *self,
-                                         const GArray *handles,
-                                         GArray **ret,
-                                         GError **error)
-{
-  return gabble_group_mixin_get_handle_owners (G_OBJECT (self), handles, ret,
-      error);
+  tp_svc_channel_return_from_get_handle (context, priv->handle_type,
+      priv->handle);
 }
 
 
@@ -626,131 +596,27 @@ gabble_roster_channel_get_handle_owners (GabbleRosterChannel *self,
  *
  * Implements D-Bus method GetInterfaces
  * on interface org.freedesktop.Telepathy.Channel
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
  */
-gboolean
-gabble_roster_channel_get_interfaces (GabbleRosterChannel *self,
-                                      gchar ***ret,
-                                      GError **error)
+static void
+gabble_roster_channel_get_interfaces (TpSvcChannel *self,
+                                      DBusGMethodInvocation *context)
 {
   const char *interfaces[] = { TP_IFACE_CHANNEL_INTERFACE_GROUP, NULL };
 
-  *ret = g_strdupv ((gchar **) interfaces);
-
-  return TRUE;
+  tp_svc_channel_return_from_get_interfaces (context, interfaces);
 }
 
 
-/**
- * gabble_roster_channel_get_local_pending_members
- *
- * Implements D-Bus method GetLocalPendingMembers
- * on interface org.freedesktop.Telepathy.Channel.Interface.Group
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
- */
-gboolean
-gabble_roster_channel_get_local_pending_members (GabbleRosterChannel *self,
-                                                 GArray **ret,
-                                                 GError **error)
+static void
+channel_iface_init(gpointer g_iface, gpointer iface_data)
 {
-  return gabble_group_mixin_get_local_pending_members (G_OBJECT (self), ret,
-      error);
+  TpSvcChannelClass *klass = (TpSvcChannelClass *)g_iface;
+
+#define IMPLEMENT(x) tp_svc_channel_implement_##x (\
+    klass, gabble_roster_channel_##x)
+  IMPLEMENT(close);
+  IMPLEMENT(get_channel_type);
+  IMPLEMENT(get_handle);
+  IMPLEMENT(get_interfaces);
+#undef IMPLEMENT
 }
-
-
-/**
- * gabble_roster_channel_get_members
- *
- * Implements D-Bus method GetMembers
- * on interface org.freedesktop.Telepathy.Channel.Interface.Group
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
- */
-gboolean
-gabble_roster_channel_get_members (GabbleRosterChannel *self,
-                                   GArray **ret,
-                                   GError **error)
-{
-  return gabble_group_mixin_get_members (G_OBJECT (self), ret, error);
-}
-
-
-/**
- * gabble_roster_channel_get_remote_pending_members
- *
- * Implements D-Bus method GetRemotePendingMembers
- * on interface org.freedesktop.Telepathy.Channel.Interface.Group
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
- */
-gboolean
-gabble_roster_channel_get_remote_pending_members (GabbleRosterChannel *self,
-                                                  GArray **ret,
-                                                  GError **error)
-{
-  return gabble_group_mixin_get_remote_pending_members (G_OBJECT (self), ret,
-      error);
-}
-
-
-/**
- * gabble_roster_channel_get_self_handle
- *
- * Implements D-Bus method GetSelfHandle
- * on interface org.freedesktop.Telepathy.Channel.Interface.Group
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
- */
-gboolean
-gabble_roster_channel_get_self_handle (GabbleRosterChannel *self,
-                                       guint *ret,
-                                       GError **error)
-{
-  return gabble_group_mixin_get_self_handle (G_OBJECT (self), ret, error);
-}
-
-
-/**
- * gabble_roster_channel_remove_members
- *
- * Implements D-Bus method RemoveMembers
- * on interface org.freedesktop.Telepathy.Channel.Interface.Group
- *
- * @error: Used to return a pointer to a GError detailing any error
- *         that occurred, D-Bus will throw the error only if this
- *         function returns FALSE.
- *
- * Returns: TRUE if successful, FALSE if an error was thrown.
- */
-gboolean
-gabble_roster_channel_remove_members (GabbleRosterChannel *self,
-                                      const GArray *contacts,
-                                      const gchar *message,
-                                      GError **error)
-{
-  return gabble_group_mixin_remove_members (G_OBJECT (self), contacts, message,
-      error);
-}
-
