@@ -22,15 +22,23 @@
 #ifndef __GABBLE_TYPES_H__
 #define __GABBLE_TYPES_H__
 
+#include <telepathy-glib/handle.h>
+
 G_BEGIN_DECLS
 
 typedef struct _GabbleConnection GabbleConnection;
 typedef struct _GabbleDisco GabbleDisco;
-typedef struct _GabbleHandleRepo GabbleHandleRepo;
+typedef struct _GabblePresence GabblePresence;
 typedef struct _GabblePresenceCache GabblePresenceCache;
 typedef struct _GabbleRoster GabbleRoster;
 typedef struct _GabbleRosterChannel GabbleRosterChannel;
-typedef guint32 GabbleHandle;
+typedef struct _GabbleVCardManager GabbleVCardManager;
+
+typedef enum {
+    INITIATOR_INVALID = -1,
+    INITIATOR_LOCAL = 0,
+    INITIATOR_REMOTE,
+} JingleInitiator;
 
 G_END_DECLS
 
