@@ -37,9 +37,9 @@
 #include "gabble-connection.h"
 #include "gabble-im-channel.h"
 #include "handles.h"
-#include "telepathy-interfaces.h"
+#include <telepathy-glib/interfaces.h>
 #include "text-mixin.h"
-#include "tp-channel-factory-iface.h"
+#include <telepathy-glib/channel-factory-iface.h>
 
 #include "im-factory.h"
 
@@ -181,7 +181,7 @@ gabble_im_factory_class_init (GabbleImFactoryClass *gabble_im_factory_class)
 
 }
 
-static GabbleIMChannel *new_im_channel (GabbleImFactory *fac, GabbleHandle handle);
+static GabbleIMChannel *new_im_channel (GabbleImFactory *fac, TpHandle handle);
 
 static void im_channel_closed_cb (GabbleIMChannel *chan, gpointer user_data);
 
@@ -199,24 +199,27 @@ im_factory_message_cb (LmMessageHandler *handler,
 {
   GabbleImFactory *fac = GABBLE_IM_FACTORY (user_data);
   GabbleImFactoryPrivate *priv = GABBLE_IM_FACTORY_GET_PRIVATE (fac);
+  TpBaseConnection *conn = (TpBaseConnection *)priv->conn;
 
   const gchar *from, *body, *body_offset;
   time_t stamp;
   TpChannelTextMessageType msgtype;
-  GabbleHandle handle;
+  TpHandle handle;
   GabbleIMChannel *chan;
-  GabbleTextMixinSendError send_error;
+  gint state;
+  TpChannelTextSendError send_error;
 
-  if (!gabble_text_mixin_parse_incoming_message (message, &from, &stamp, &msgtype, &body, &body_offset, &send_error))
+  if (!gabble_text_mixin_parse_incoming_message (message, &from, &stamp, &msgtype, &body, &body_offset, &state, &send_error))
     return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
 
-  if (body == NULL)
+  if (body == NULL && state == -1)
     {
       NODE_DEBUG (message->node, "got a message without a body field, ignoring");
       return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
     }
 
-  handle = gabble_handle_for_contact (priv->conn->handles, from, FALSE);
+  handle = gabble_handle_for_contact (
+      conn->handles[TP_HANDLE_TYPE_CONTACT], from, FALSE);
   if (handle == 0)
     {
       NODE_DEBUG (message->node, "ignoring message node from malformed jid");
@@ -230,7 +233,7 @@ im_factory_message_cb (LmMessageHandler *handler,
 
   if (chan == NULL)
     {
-      if (send_error != CHANNEL_TEXT_SEND_NO_ERROR)
+      if (send_error != TP_CHANNEL_SEND_NO_ERROR)
         {
           DEBUG ("ignoring message error; no sending channel");
           return LM_HANDLER_RESULT_REMOVE_MESSAGE;
@@ -241,14 +244,17 @@ im_factory_message_cb (LmMessageHandler *handler,
       chan = new_im_channel (fac, handle);
     }
 
-  if (send_error != CHANNEL_TEXT_SEND_NO_ERROR)
+  if (send_error != TP_CHANNEL_SEND_NO_ERROR)
     {
-      _gabble_text_mixin_send_error_signal (G_OBJECT (chan), send_error, stamp,
-          msgtype, body_offset);
+      tp_svc_channel_type_text_emit_send_error ((TpSvcChannelTypeText *)chan,
+          send_error, stamp, msgtype, body_offset);
       return LM_HANDLER_RESULT_REMOVE_MESSAGE;
     }
 
-  if (_gabble_im_channel_receive (chan, msgtype, handle, from,
+  if (state != -1)
+    _gabble_im_channel_state_receive (chan, state);
+
+  if (body != NULL && _gabble_im_channel_receive (chan, msgtype, handle, from,
                                   stamp, body_offset))
     return LM_HANDLER_RESULT_REMOVE_MESSAGE;
 
@@ -266,7 +272,7 @@ im_channel_closed_cb (GabbleIMChannel *chan, gpointer user_data)
 {
   GabbleImFactory *conn = GABBLE_IM_FACTORY (user_data);
   GabbleImFactoryPrivate *priv = GABBLE_IM_FACTORY_GET_PRIVATE (conn);
-  GabbleHandle contact_handle;
+  TpHandle contact_handle;
 
   if (priv->channels)
     {
@@ -282,17 +288,20 @@ im_channel_closed_cb (GabbleIMChannel *chan, gpointer user_data)
  * new_im_channel
  */
 static GabbleIMChannel *
-new_im_channel (GabbleImFactory *fac, GabbleHandle handle)
+new_im_channel (GabbleImFactory *fac, TpHandle handle)
 {
   GabbleImFactoryPrivate *priv;
+  TpBaseConnection *conn;
   GabbleIMChannel *chan;
   char *object_path;
 
   g_assert (GABBLE_IS_IM_FACTORY (fac));
 
   priv = GABBLE_IM_FACTORY_GET_PRIVATE (fac);
+  conn = (TpBaseConnection *)priv->conn;
 
-  object_path = g_strdup_printf ("%s/ImChannel%u", priv->conn->object_path, handle);
+  object_path = g_strdup_printf ("%s/ImChannel%u",
+      conn->object_path, handle);
 
   chan = g_object_new (GABBLE_TYPE_IM_CHANNEL,
                        "connection", priv->conn,
@@ -306,7 +315,7 @@ new_im_channel (GabbleImFactory *fac, GabbleHandle handle)
 
   g_hash_table_insert (priv->channels, GINT_TO_POINTER (handle), chan);
 
-  g_signal_emit_by_name (fac, "new-channel", chan);
+  tp_channel_factory_iface_emit_new_channel (fac, (TpChannelIface *)chan, NULL);
 
   g_free (object_path);
 
@@ -402,12 +411,15 @@ gabble_im_factory_iface_request (TpChannelFactoryIface *iface,
                                  const gchar *chan_type,
                                  TpHandleType handle_type,
                                  guint handle,
-                                 TpChannelIface **ret)
+                                 gpointer request,
+                                 TpChannelIface **ret,
+                                 GError **error)
 {
   GabbleImFactory *fac = GABBLE_IM_FACTORY (iface);
   GabbleImFactoryPrivate *priv = GABBLE_IM_FACTORY_GET_PRIVATE (fac);
+  TpBaseConnection *conn = (TpBaseConnection *)priv->conn;
   GabbleIMChannel *chan;
-  GError *error;
+  TpChannelFactoryRequestStatus status;
 
   if (strcmp (chan_type, TP_IFACE_CHANNEL_TYPE_TEXT))
     return TP_CHANNEL_FACTORY_REQUEST_STATUS_NOT_IMPLEMENTED;
@@ -415,18 +427,22 @@ gabble_im_factory_iface_request (TpChannelFactoryIface *iface,
   if (handle_type != TP_HANDLE_TYPE_CONTACT)
     return TP_CHANNEL_FACTORY_REQUEST_STATUS_NOT_AVAILABLE;
 
-  if (!gabble_handle_is_valid (priv->conn->handles, handle_type, handle, &error))
-    return TP_CHANNEL_FACTORY_REQUEST_STATUS_INVALID_HANDLE;
+  if (!tp_handle_is_valid (conn->handles[TP_HANDLE_TYPE_CONTACT],
+        handle, error))
+    return TP_CHANNEL_FACTORY_REQUEST_STATUS_ERROR;
 
   chan = g_hash_table_lookup (priv->channels, GINT_TO_POINTER (handle));
+
+  status = TP_CHANNEL_FACTORY_REQUEST_STATUS_EXISTING;
   if (!chan)
     {
+      status = TP_CHANNEL_FACTORY_REQUEST_STATUS_CREATED;
       chan = new_im_channel (fac, handle);
-      if (!chan) return TP_CHANNEL_FACTORY_REQUEST_STATUS_INVALID_HANDLE;
     }
 
+  g_assert (chan);
   *ret = TP_CHANNEL_IFACE (chan);
-  return TP_CHANNEL_FACTORY_REQUEST_STATUS_DONE;
+  return status;
 }
 
 static void
