@@ -38,9 +38,9 @@
 #include "handles.h"
 #include "media-factory.h"
 #include "namespaces.h"
-#include "telepathy-interfaces.h"
+#include <telepathy-glib/interfaces.h>
 #include "text-mixin.h"
-#include "tp-channel-factory-iface.h"
+#include <telepathy-glib/channel-factory-iface.h>
 #include "util.h"
 
 static void gabble_media_factory_iface_init (gpointer g_iface, gpointer iface_data);
@@ -61,12 +61,18 @@ struct _GabbleMediaFactoryPrivate
 {
   GabbleConnection *conn;
   LmMessageHandler *jingle_cb;
+  LmMessageHandler *jingle_info_cb;
 
   GPtrArray *channels;
   guint channel_index;
 
   GHashTable *session_chans;
 
+  gboolean get_stun_from_jingle;
+  gchar *stun_server;
+  guint16 stun_port;
+  gchar *relay_token;
+
   gboolean dispose_has_run;
 };
 
@@ -83,6 +89,7 @@ gabble_media_factory_init (GabbleMediaFactory *fac)
   priv->channel_index = 0;
 
   priv->jingle_cb = NULL;
+  priv->jingle_info_cb = NULL;
 
   priv->conn = NULL;
   priv->dispose_has_run = FALSE;
@@ -118,6 +125,9 @@ gabble_media_factory_dispose (GObject *object)
   DEBUG ("dispose called");
   priv->dispose_has_run = TRUE;
 
+  g_assert (priv->jingle_cb == NULL);
+  g_assert (priv->jingle_info_cb == NULL);
+
   tp_channel_factory_iface_close_all (TP_CHANNEL_FACTORY_IFACE (object));
   g_assert (priv->channels == NULL);
 
@@ -128,6 +138,9 @@ gabble_media_factory_dispose (GObject *object)
       priv->session_chans = NULL;
     }
 
+  g_free (priv->stun_server);
+  g_free (priv->relay_token);
+
   if (G_OBJECT_CLASS (gabble_media_factory_parent_class)->dispose)
     G_OBJECT_CLASS (gabble_media_factory_parent_class)->dispose (object);
 }
@@ -196,7 +209,8 @@ gabble_media_factory_class_init (GabbleMediaFactoryClass *gabble_media_factory_c
 
 }
 
-static GabbleMediaChannel *new_media_channel (GabbleMediaFactory *fac, GabbleHandle handle);
+static gboolean _gabble_media_factory_sid_in_use (GabbleMediaFactory *fac, const gchar *sid);
+static GabbleMediaChannel *new_media_channel (GabbleMediaFactory *fac, TpHandle handle);
 static void media_channel_closed_cb (GabbleMediaChannel *chan, gpointer user_data);
 
 /**
@@ -214,11 +228,14 @@ media_factory_jingle_cb (LmMessageHandler *handler,
 {
   GabbleMediaFactory *fac = GABBLE_MEDIA_FACTORY (user_data);
   GabbleMediaFactoryPrivate *priv = GABBLE_MEDIA_FACTORY_GET_PRIVATE (fac);
+  TpBaseConnection *conn = (TpBaseConnection *)priv->conn;
   LmMessageNode *iq_node, *session_node;
   const gchar *from, *id, *action, *sid;
   gchar *resource;
-  GabbleHandle handle;
+  TpHandle handle;
   GabbleMediaChannel *chan = NULL;
+  gboolean chan_is_new = FALSE;
+  GError *error = NULL;
 
   g_assert (lmconn == priv->conn->lmconn);
 
@@ -259,7 +276,8 @@ media_factory_jingle_cb (LmMessageHandler *handler,
       goto BAD_REQUEST;
     }
 
-  handle = gabble_handle_for_contact (priv->conn->handles, from, FALSE);
+  handle = gabble_handle_for_contact (
+      conn->handles[TP_HANDLE_TYPE_CONTACT], from, FALSE);
   if (handle == 0)
     {
       NODE_DEBUG (iq_node, "unable to get handle for sender");
@@ -294,15 +312,18 @@ media_factory_jingle_cb (LmMessageHandler *handler,
   if (chan == NULL)
     {
       /* if the session is unknown, the only allowed action is "initiate" */
-      if (g_strdiff (action, "initiate") && g_strdiff (action, "session-initiate"))
+      if (tp_strdiff (action, "initiate") &&
+          tp_strdiff (action, "session-initiate"))
         {
-          NODE_DEBUG (iq_node, "action is not \"initiate\" or \"session-initiate\", rejecting");
+          NODE_DEBUG (iq_node,
+              "action is not \"initiate\" or \"session-initiate\", rejecting");
           goto BAD_REQUEST;
         }
 
       DEBUG ("creating media channel");
 
       chan = new_media_channel (fac, handle);
+      chan_is_new = TRUE;
     }
 
   g_assert (chan != NULL);
@@ -310,15 +331,32 @@ media_factory_jingle_cb (LmMessageHandler *handler,
   DEBUG ("dispatching to session %s", sid);
   g_object_ref (chan);
   gabble_decode_jid (from, NULL, NULL, &resource);
-  _gabble_media_channel_dispatch_session_action (chan, handle, resource,
-      sid, message, session_node, action);
+
+  if (_gabble_media_channel_dispatch_session_action (chan, handle, resource,
+        sid, message, session_node, action, &error))
+    {
+      if (chan_is_new)
+        tp_channel_factory_iface_emit_new_channel (fac,
+            (TpChannelIface *)chan, NULL);
+    }
+  else
+    {
+      if (chan_is_new)
+        gabble_media_channel_close (chan);
+
+      g_assert (error != NULL);
+      _gabble_connection_send_iq_error (priv->conn, message, error->code,
+          error->message);
+    }
+
   g_object_unref (chan);
   g_free (resource);
 
   return LM_HANDLER_RESULT_REMOVE_MESSAGE;
 
 BAD_REQUEST:
-  _gabble_connection_send_iq_error (priv->conn, message, XMPP_ERROR_BAD_REQUEST);
+  _gabble_connection_send_iq_error (
+    priv->conn, message, XMPP_ERROR_BAD_REQUEST, NULL);
 
   return LM_HANDLER_RESULT_REMOVE_MESSAGE;
 }
@@ -346,7 +384,7 @@ _gabble_media_factory_get_unique_sid (GabbleMediaFactory *fac)
   return sid;
 }
 
-gboolean
+static gboolean
 _gabble_media_factory_sid_in_use (GabbleMediaFactory *fac, const gchar *sid)
 {
   GabbleMediaFactoryPrivate *priv = GABBLE_MEDIA_FACTORY_GET_PRIVATE (fac);
@@ -432,18 +470,20 @@ media_channel_closed_cb (GabbleMediaChannel *chan, gpointer user_data)
  * Creates a new empty GabbleMediaChannel.
  */
 static GabbleMediaChannel *
-new_media_channel (GabbleMediaFactory *fac, GabbleHandle creator)
+new_media_channel (GabbleMediaFactory *fac, TpHandle creator)
 {
   GabbleMediaFactoryPrivate *priv;
+  TpBaseConnection *conn;
   GabbleMediaChannel *chan;
   gchar *object_path;
 
   g_assert (GABBLE_IS_MEDIA_FACTORY (fac));
 
   priv = GABBLE_MEDIA_FACTORY_GET_PRIVATE (fac);
+  conn = (TpBaseConnection *)priv->conn;
 
-  object_path = g_strdup_printf ("%s/MediaChannel%u", priv->conn->object_path,
-                                 priv->channel_index);
+  object_path = g_strdup_printf ("%s/MediaChannel%u",
+      conn->object_path, priv->channel_index);
   priv->channel_index += 1;
 
   chan = g_object_new (GABBLE_TYPE_MEDIA_CHANNEL,
@@ -453,86 +493,167 @@ new_media_channel (GabbleMediaFactory *fac, GabbleHandle creator)
                        "creator", creator,
                        NULL);
 
+  if (priv->stun_server != NULL)
+    {
+      g_object_set ((GObject *) chan, "stun-server", priv->stun_server, NULL);
+
+      if (priv->stun_port != 0)
+        g_object_set ((GObject *) chan, "stun-port", priv->stun_port, NULL);
+    }
+
+  if (priv->relay_token != NULL)
+    {
+      g_object_set ((GObject *) chan, "gtalk-p2p-relay-token", priv->relay_token, NULL);
+    }
+
   DEBUG ("object path %s", object_path);
 
   g_signal_connect (chan, "closed", (GCallback) media_channel_closed_cb, fac);
 
   g_ptr_array_add (priv->channels, chan);
 
-  g_signal_emit_by_name (fac, "new-channel", chan);
-
   g_free (object_path);
 
   return chan;
 }
 
 
-static GabbleMediaChannel *
-find_media_channel_with_handle (GabbleMediaFactory *fac, GabbleHandle handle)
+static void
+jingle_info_send_request (GabbleMediaFactory *fac)
 {
   GabbleMediaFactoryPrivate *priv = GABBLE_MEDIA_FACTORY_GET_PRIVATE (fac);
-  guint i, j;
+  TpBaseConnection *base = (TpBaseConnection *) priv->conn;
+  LmMessage *msg;
+  LmMessageNode *node;
+  const gchar *jid;
+  GError *error = NULL;
 
-  for (i = 0; i < priv->channels->len; i++)
+  jid = tp_handle_inspect (base->handles[TP_HANDLE_TYPE_CONTACT],
+      base->self_handle);
+  msg = lm_message_new_with_sub_type (jid, LM_MESSAGE_TYPE_IQ,
+      LM_MESSAGE_SUB_TYPE_GET);
+
+  node = lm_message_node_add_child (msg->node, "query", NULL);
+  lm_message_node_set_attribute (node, "xmlns", NS_GOOGLE_JINGLE_INFO);
+
+  if (!_gabble_connection_send (priv->conn, msg, &error))
     {
-      GArray *arr;
-      GError *err;
-
-      GabbleMediaChannel *chan = g_ptr_array_index (priv->channels, i);
-
-      /* search members */
-      if (!gabble_group_mixin_get_members (G_OBJECT (chan), &arr, &err))
-        {
-          DEBUG ("get_members failed: %s", err->message);
-          g_error_free (err);
-          continue;
-        }
-
-      for (j = 0; j < arr->len; j++)
-        if (g_array_index (arr, guint32, i) == handle)
-          {
-            g_array_free (arr, TRUE);
-            return chan;
-          }
-
-      g_array_free (arr, TRUE);
-
-      /* search local pending */
-      if (!gabble_group_mixin_get_local_pending_members (G_OBJECT (chan), &arr, &err))
-        {
-          DEBUG ("get_local_pending_members failed: %s", err->message);
-          g_error_free (err);
-          continue;
-        }
-
-      for (j = 0; j < arr->len; j++)
-        if (g_array_index (arr, guint32, i) == handle)
-          {
-            g_array_free (arr, TRUE);
-            return chan;
-          }
-
-      g_array_free (arr, TRUE);
-
-      /* search remote pending */
-      if (!gabble_group_mixin_get_remote_pending_members (G_OBJECT (chan), &arr, &err))
-        {
-          DEBUG ("get_remote_pending_members failed: %s", err->message);
-          g_error_free (err);
-          continue;
-        }
-
-      for (j = 0; j < arr->len; j++)
-        if (g_array_index (arr, guint32, i) == handle)
-          {
-            g_array_free (arr, TRUE);
-            return chan;
-          }
-
-      g_array_free (arr, TRUE);
+      DEBUG ("jingle info send failed: %s\n", error->message);
+      g_error_free (error);
     }
 
-  return NULL;
+  lm_message_unref (msg);
+}
+
+
+/**
+ * jingle_info_iq_callback
+ *
+ * Called by loudmouth when we get an incoming <iq>. This handler
+ * is concerned only with Jingle info queries.
+ */
+LmHandlerResult
+jingle_info_iq_callback (LmMessageHandler *handler,
+                         LmConnection *lmconn,
+                         LmMessage *message,
+                         gpointer user_data)
+{
+  GabbleMediaFactory *fac = GABBLE_MEDIA_FACTORY (user_data);
+  GabbleMediaFactoryPrivate *priv = GABBLE_MEDIA_FACTORY_GET_PRIVATE (fac);
+  LmMessageSubType sub_type;
+  LmMessageNode *query_node, *node;
+
+  query_node = lm_message_node_get_child_with_namespace (message->node,
+      "query", NS_GOOGLE_JINGLE_INFO);
+
+  if (query_node == NULL)
+    return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
+
+  sub_type = lm_message_get_sub_type (message);
+
+  if (sub_type == LM_MESSAGE_SUB_TYPE_ERROR)
+    {
+      GabbleXmppError xmpp_error = INVALID_XMPP_ERROR;
+
+      node = lm_message_node_get_child (message->node, "error");
+      if (node != NULL)
+        {
+          xmpp_error = gabble_xmpp_error_from_node (node);
+        }
+
+      DEBUG ("jingle info error: %s", xmpp_error == INVALID_XMPP_ERROR ?
+          "unknown error" : gabble_xmpp_error_string (xmpp_error));
+
+      return LM_HANDLER_RESULT_REMOVE_MESSAGE;
+    }
+
+  if (sub_type != LM_MESSAGE_SUB_TYPE_RESULT &&
+      sub_type != LM_MESSAGE_SUB_TYPE_SET)
+    {
+      DEBUG ("jingle info: unexpected IQ type, ignoring");
+
+      return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
+    }
+
+  if (priv->get_stun_from_jingle)
+    node = lm_message_node_get_child (query_node, "stun");
+  else
+    node = NULL;
+
+  if (node != NULL)
+    {
+      node = lm_message_node_get_child (node, "server");
+
+      if (node != NULL)
+        {
+          const gchar *server;
+          const gchar *port;
+
+          server = lm_message_node_get_attribute (node, "host");
+          port = lm_message_node_get_attribute (node, "udp");
+
+          if (server != NULL)
+            {
+              DEBUG ("jingle info: got stun server %s", server);
+              g_free (priv->stun_server);
+              priv->stun_server = g_strdup (server);
+            }
+
+          if (port != NULL)
+            {
+              DEBUG ("jingle info: got stun port %s", port);
+              priv->stun_port = atoi (port);
+            }
+        }
+    }
+
+  node = lm_message_node_get_child (query_node, "relay");
+
+  if (node != NULL)
+    {
+      node = lm_message_node_get_child (node, "token");
+
+      if (node != NULL)
+        {
+          const gchar *token;
+
+          token = lm_message_node_get_value (node);
+
+          if (token != NULL)
+            {
+              DEBUG ("jingle info: got relay token %s", token);
+              g_free (priv->relay_token);
+              priv->relay_token = g_strdup (token);
+            }
+        }
+    }
+
+  if (sub_type == LM_MESSAGE_SUB_TYPE_SET)
+    {
+      _gabble_connection_acknowledge_set_iq (priv->conn, message);
+    }
+
+  return LM_HANDLER_RESULT_REMOVE_MESSAGE;
 }
 
 
@@ -549,7 +670,7 @@ gabble_media_factory_iface_close_all (TpChannelFactoryIface *iface)
       GPtrArray *tmp = priv->channels;
       priv->channels = NULL;
 
-      int i;
+      guint i;
 
       for (i = 0; i < tmp->len; i++)
         {
@@ -582,16 +703,50 @@ gabble_media_factory_iface_connecting (TpChannelFactoryIface *iface)
 
   DEBUG ("adding callbacks");
 
+  g_assert (priv->jingle_cb == NULL);
+  g_assert (priv->jingle_info_cb == NULL);
+
   priv->jingle_cb = lm_message_handler_new (media_factory_jingle_cb, fac, NULL);
   lm_connection_register_message_handler (priv->conn->lmconn, priv->jingle_cb,
                                           LM_MESSAGE_TYPE_IQ,
                                           LM_HANDLER_PRIORITY_NORMAL);
+
+  priv->jingle_info_cb = lm_message_handler_new (jingle_info_iq_callback, fac,
+      NULL);
+  lm_connection_register_message_handler (priv->conn->lmconn,
+                                          priv->jingle_info_cb,
+                                          LM_MESSAGE_TYPE_IQ,
+                                          LM_HANDLER_PRIORITY_NORMAL);
 }
 
 static void
 gabble_media_factory_iface_connected (TpChannelFactoryIface *iface)
 {
-  /* nothing to do */
+  GabbleMediaFactory *fac = GABBLE_MEDIA_FACTORY (iface);
+  GabbleMediaFactoryPrivate *priv = GABBLE_MEDIA_FACTORY_GET_PRIVATE (fac);
+  gchar *stun_server = NULL;
+  guint stun_port = 0;
+
+  g_object_get (priv->conn,
+      "stun-server", &stun_server,
+      "stun-port", &stun_port,
+      NULL);
+
+  if (stun_server == NULL)
+    {
+      priv->get_stun_from_jingle = TRUE;
+    }
+  else
+    {
+      g_free (priv->stun_server);
+      priv->stun_server = stun_server;
+      priv->stun_port = stun_port;
+    }
+
+  if (priv->conn->features & GABBLE_CONNECTION_FEATURES_GOOGLE_JINGLE_INFO)
+    {
+      jingle_info_send_request (fac);
+    }
 }
 
 static void
@@ -600,14 +755,21 @@ gabble_media_factory_iface_disconnected (TpChannelFactoryIface *iface)
   GabbleMediaFactory *fac = GABBLE_MEDIA_FACTORY (iface);
   GabbleMediaFactoryPrivate *priv = GABBLE_MEDIA_FACTORY_GET_PRIVATE (fac);
 
-  g_assert (priv->jingle_cb != NULL);
-
   DEBUG ("removing callbacks");
 
+  g_assert (priv->jingle_cb != NULL);
+  g_assert (priv->jingle_info_cb != NULL);
+
   lm_connection_unregister_message_handler (priv->conn->lmconn, priv->jingle_cb,
                                             LM_MESSAGE_TYPE_IQ);
   lm_message_handler_unref (priv->jingle_cb);
   priv->jingle_cb = NULL;
+
+  lm_connection_unregister_message_handler (priv->conn->lmconn, priv->jingle_info_cb,
+                                            LM_MESSAGE_TYPE_IQ);
+  lm_message_handler_unref (priv->jingle_info_cb);
+  priv->jingle_info_cb = NULL;
+
 }
 
 static void
@@ -615,7 +777,7 @@ gabble_media_factory_iface_foreach (TpChannelFactoryIface *iface, TpChannelFunc
 {
   GabbleMediaFactory *fac = GABBLE_MEDIA_FACTORY (iface);
   GabbleMediaFactoryPrivate *priv = GABBLE_MEDIA_FACTORY_GET_PRIVATE (fac);
-  int i;
+  guint i;
 
   for (i = 0; i < priv->channels->len; i++)
     {
@@ -628,12 +790,14 @@ gabble_media_factory_iface_request (TpChannelFactoryIface *iface,
                                     const gchar *chan_type,
                                     TpHandleType handle_type,
                                     guint handle,
-                                    TpChannelIface **ret)
+                                    gpointer request,
+                                    TpChannelIface **ret,
+                                    GError **error)
 {
   GabbleMediaFactory *fac = GABBLE_MEDIA_FACTORY (iface);
   GabbleMediaFactoryPrivate *priv = GABBLE_MEDIA_FACTORY_GET_PRIVATE (fac);
+  TpBaseConnection *conn = (TpBaseConnection *)priv->conn;
   GabbleMediaChannel *chan = NULL;
-  GError *error;
 
   if (strcmp (chan_type, TP_IFACE_CHANNEL_TYPE_STREAMED_MEDIA))
     return TP_CHANNEL_FACTORY_REQUEST_STATUS_NOT_IMPLEMENTED;
@@ -641,44 +805,31 @@ gabble_media_factory_iface_request (TpChannelFactoryIface *iface,
   if (handle_type == 0)
     {
       /* create an empty channel */
-      chan = new_media_channel (fac, priv->conn->self_handle);
+      chan = new_media_channel (fac, conn->self_handle);
+    }
+  else if (handle_type == TP_HANDLE_TYPE_CONTACT)
+    {
+      chan = new_media_channel (fac, conn->self_handle);
+
+      if (!_gabble_media_channel_add_member (
+            TP_SVC_CHANNEL_INTERFACE_GROUP (chan), handle, "", error))
+        {
+          gabble_media_channel_close (chan);
+
+          return TP_CHANNEL_FACTORY_REQUEST_STATUS_ERROR;
+        }
     }
   else
     {
-      if (handle_type != TP_HANDLE_TYPE_CONTACT)
-      {
-        return TP_CHANNEL_FACTORY_REQUEST_STATUS_INVALID_HANDLE;
-      }
-
-      /* have we already got a channel with this handle? */
-      chan = find_media_channel_with_handle (fac, handle);
-
-      /* no: create it and add the peer to it */
-      if (!chan)
-        {
-          GArray *members;
-          gboolean ret;
-
-          chan = new_media_channel (fac, priv->conn->self_handle);
-
-          members = g_array_sized_new (FALSE, FALSE, sizeof (GabbleHandle), 1);
-          g_array_append_val (members, handle);
-
-          ret = gabble_group_mixin_add_members (G_OBJECT (chan), members, "", &error);
-
-          g_array_free (members, TRUE);
-
-          if (!ret)
-            {
-              g_error_free (error);
-              gabble_media_channel_close (chan, NULL);
-              return TP_CHANNEL_FACTORY_REQUEST_STATUS_INVALID_HANDLE;
-            }
-        }
+      return TP_CHANNEL_FACTORY_REQUEST_STATUS_INVALID_HANDLE;
     }
 
+  g_assert (chan != NULL);
+  tp_channel_factory_iface_emit_new_channel (fac, (TpChannelIface *)chan,
+      request);
+
   *ret = TP_CHANNEL_IFACE (chan);
-  return TP_CHANNEL_FACTORY_REQUEST_STATUS_DONE;
+  return TP_CHANNEL_FACTORY_REQUEST_STATUS_CREATED;
 }
 
 static void
