@@ -35,14 +35,15 @@
 #include "debug.h"
 #include "disco.h"
 #include "gabble-connection.h"
-#include "gabble-presence-cache.h"
+#include "presence-cache.h"
 #include "gabble-muc-channel.h"
+#include "gabble-roomlist-channel.h"
 #include "handles.h"
 #include "muc-factory.h"
 #include "namespaces.h"
-#include "telepathy-interfaces.h"
+#include <telepathy-glib/interfaces.h>
 #include "text-mixin.h"
-#include "tp-channel-factory-iface.h"
+#include <telepathy-glib/channel-factory-iface.h>
 #include "util.h"
 
 static void gabble_muc_factory_iface_init (gpointer g_iface, gpointer iface_data);
@@ -66,6 +67,8 @@ struct _GabbleMucFactoryPrivate
   LmMessageHandler *presence_cb;
 
   GHashTable *channels;
+  GabbleRoomlistChannel *roomlist_channel;
+
   GHashTable *disco_requests;
 
   gboolean dispose_has_run;
@@ -205,12 +208,15 @@ static GabbleMucChannel *
 get_muc_from_jid (GabbleMucFactory *fac, const gchar *jid)
 {
   GabbleMucFactoryPrivate *priv = GABBLE_MUC_FACTORY_GET_PRIVATE (fac);
-  GabbleHandle handle;
+  TpBaseConnection *conn = (TpBaseConnection *)priv->conn;
+  TpHandle handle;
   GabbleMucChannel *chan = NULL;
 
-  if (gabble_handle_for_room_exists (priv->conn->handles, jid, TRUE))
+  if (gabble_handle_for_room_exists (
+        conn->handles[TP_HANDLE_TYPE_ROOM], jid, TRUE))
     {
-      handle = gabble_handle_for_room (priv->conn->handles, jid);
+      handle = gabble_handle_for_room (
+          conn->handles[TP_HANDLE_TYPE_ROOM], jid);
 
       chan = g_hash_table_lookup (priv->channels, GUINT_TO_POINTER (handle));
     }
@@ -230,9 +236,9 @@ muc_channel_closed_cb (GabbleMucChannel *chan, gpointer user_data)
 {
   GabbleMucFactory *fac = GABBLE_MUC_FACTORY (user_data);
   GabbleMucFactoryPrivate *priv = GABBLE_MUC_FACTORY_GET_PRIVATE (fac);
-  GabbleHandle room_handle;
+  TpHandle room_handle;
 
-  if (priv->channels)
+  if (priv->channels != NULL)
     {
       g_object_get (chan, "handle", &room_handle, NULL);
 
@@ -250,7 +256,7 @@ muc_ready_cb (GabbleMucChannel *chan,
 
   DEBUG ("chan=%p", chan);
 
-  g_signal_emit_by_name (fac, "new-channel", chan);
+  tp_channel_factory_iface_emit_new_channel (fac, (TpChannelIface *)chan, NULL);
 }
 
 static void
@@ -262,22 +268,25 @@ muc_join_error_cb (GabbleMucChannel *chan,
 
   DEBUG ("error->code=%u, error->message=\"%s\"", error->code, error->message);
 
-  g_signal_emit_by_name (fac, "channel-error", chan, error);
+  tp_channel_factory_iface_emit_channel_error (fac, (TpChannelIface *)chan,
+      error, NULL);
 }
 
 /**
  * new_muc_channel
  */
 static GabbleMucChannel *
-new_muc_channel (GabbleMucFactory *fac, GabbleHandle handle)
+new_muc_channel (GabbleMucFactory *fac, TpHandle handle, gboolean invite_self)
 {
   GabbleMucFactoryPrivate *priv = GABBLE_MUC_FACTORY_GET_PRIVATE (fac);
+  TpBaseConnection *conn = (TpBaseConnection *)priv->conn;
   GabbleMucChannel *chan;
   char *object_path;
 
   g_assert (g_hash_table_lookup (priv->channels, GINT_TO_POINTER (handle)) == NULL);
 
-  object_path = g_strdup_printf ("%s/MucChannel%u", priv->conn->object_path, handle);
+  object_path = g_strdup_printf ("%s/MucChannel%u",
+      conn->object_path, handle);
 
   DEBUG ("creating new chan, object path %s", object_path);
 
@@ -285,6 +294,7 @@ new_muc_channel (GabbleMucFactory *fac, GabbleHandle handle)
                        "connection", priv->conn,
                        "object-path", object_path,
                        "handle", handle,
+                       "invite-self", invite_self,
                        NULL);
 
   g_signal_connect (chan, "closed", (GCallback) muc_channel_closed_cb, fac);
@@ -304,7 +314,7 @@ new_muc_channel (GabbleMucFactory *fac, GabbleHandle handle)
 struct DiscoInviteData {
     GabbleMucFactory *factory;
     gchar *reason;
-    GabbleHandle inviter;
+    TpHandle inviter;
 };
 
 /**
@@ -313,21 +323,23 @@ struct DiscoInviteData {
  * Callback for disco request we fired upon encountering obsolete disco.
  * If the object is in fact MUC room, create a channel for it.
  */
-void obsolete_invite_disco_cb (GabbleDisco *self,
-                               GabbleDiscoRequest *request,
-                               const gchar *jid,
-                               const gchar *node,
-                               LmMessageNode *query_result,
-                               GError* error,
-                               gpointer user_data)
+static void
+obsolete_invite_disco_cb (GabbleDisco *self,
+                          GabbleDiscoRequest *request,
+                          const gchar *jid,
+                          const gchar *node,
+                          LmMessageNode *query_result,
+                          GError* error,
+                          gpointer user_data)
 {
   struct DiscoInviteData *data = (struct DiscoInviteData *) user_data;
 
   GabbleMucFactory *fac = GABBLE_MUC_FACTORY (data->factory);
   GabbleMucFactoryPrivate *priv = GABBLE_MUC_FACTORY_GET_PRIVATE (fac);
+  TpBaseConnection *conn = (TpBaseConnection *)priv->conn;
   LmMessageNode *identity;
   const char *category, *type;
-  GabbleHandle handle;
+  TpHandle handle;
 
   g_hash_table_remove (priv->disco_requests, request);
 
@@ -346,25 +358,26 @@ void obsolete_invite_disco_cb (GabbleDisco *self,
   if (0 != strcmp (category, "conference") ||
       0 != strcmp (type, "text"))
     {
-      g_debug ("obsolete invite request specified invalid jid '%s', ignoring", jid);
+      DEBUG ("obsolete invite request specified invalid jid '%s', ignoring", jid);
     }
 
   /* OK, it's MUC after all, create a new channel */
-  handle = gabble_handle_for_room (priv->conn->handles, jid);
+  handle = gabble_handle_for_room (
+      conn->handles[TP_HANDLE_TYPE_ROOM], jid);
 
   if (g_hash_table_lookup (priv->channels, GINT_TO_POINTER (handle)) == NULL)
     {
       GabbleMucChannel *chan;
-      chan = new_muc_channel (fac, handle);
+      chan = new_muc_channel (fac, handle, FALSE);
       _gabble_muc_channel_handle_invited (chan, data->inviter, data->reason);
     }
   else
     {
-      g_debug ("ignoring invite to a room '%s' we're already in", jid);
+      DEBUG ("ignoring invite to a room '%s' we're already in", jid);
     }
 
   g_free (data->reason);
-  g_free (data);
+  g_slice_free (struct DiscoInviteData, data);
 }
 
 
@@ -382,34 +395,36 @@ muc_factory_message_cb (LmMessageHandler *handler,
 {
   GabbleMucFactory *fac = GABBLE_MUC_FACTORY (user_data);
   GabbleMucFactoryPrivate *priv = GABBLE_MUC_FACTORY_GET_PRIVATE (fac);
+  TpBaseConnection *conn = (TpBaseConnection *)priv->conn;
 
   const gchar *from, *body, *body_offset;
   time_t stamp;
   TpChannelTextMessageType msgtype;
   LmMessageNode *node;
   TpHandleType handle_type;
-  GabbleHandle room_handle, handle;
+  TpHandle room_handle, handle;
   GabbleMucChannel *chan;
-  GabbleTextMixinSendError send_error;
+  gint state;
+  TpChannelTextSendError send_error;
 
   if (!gabble_text_mixin_parse_incoming_message (message, &from, &stamp,
-        &msgtype, &body, &body_offset, &send_error))
+        &msgtype, &body, &body_offset, &state, &send_error))
     return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
 
   /* does it have a muc subnode? */
   node = lm_message_node_get_child_with_namespace (message->node, "x",
       NS_MUC_USER);
-  if (node)
+  if (node != NULL)
     {
       /* and an invitation? */
       node = lm_message_node_get_child (node, "invite");
-      if (node)
+      if (node != NULL)
         {
           LmMessageNode *reason_node;
           const gchar *invite_from, *reason;
-          GabbleHandle inviter_handle;
+          TpHandle inviter_handle;
 
-          if (send_error != CHANNEL_TEXT_SEND_NO_ERROR)
+          if (send_error != TP_CHANNEL_SEND_NO_ERROR)
             {
               NODE_DEBUG (message->node, "got a MUC invitation message "
                              "with a send error; ignoring");
@@ -427,11 +442,12 @@ muc_factory_message_cb (LmMessageHandler *handler,
               return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
             }
 
-          inviter_handle = gabble_handle_for_contact (priv->conn->handles,
-                                                      invite_from, FALSE);
+          inviter_handle = gabble_handle_for_contact (
+              conn->handles[TP_HANDLE_TYPE_CONTACT],
+              invite_from, FALSE);
 
           reason_node = lm_message_node_get_child (node, "reason");
-          if (reason_node)
+          if (reason_node != NULL)
             {
               reason = lm_message_node_get_value (reason_node);
             }
@@ -442,11 +458,12 @@ muc_factory_message_cb (LmMessageHandler *handler,
             }
 
           /* create the channel */
-          handle = gabble_handle_for_room (priv->conn->handles, from);
+          handle = gabble_handle_for_room (
+              conn->handles[TP_HANDLE_TYPE_ROOM], from);
 
           if (g_hash_table_lookup (priv->channels, GINT_TO_POINTER (handle)) == NULL)
             {
-              chan = new_muc_channel (fac, handle);
+              chan = new_muc_channel (fac, handle, FALSE);
               _gabble_muc_channel_handle_invited (chan, inviter_handle, reason);
             }
           else
@@ -459,62 +476,61 @@ muc_factory_message_cb (LmMessageHandler *handler,
     }
   else
     {
-      GabbleHandle inviter_handle;
+      TpHandle inviter_handle;
       GabbleDiscoRequest *request;
       const gchar *reason;
       struct DiscoInviteData *disco_udata;
 
       /* check for obsolete invite method */
-      for (node = message->node->children; node; node = node->next)
-        if (strcmp (node->name, "x") == 0)
-          if (lm_message_node_has_namespace (node, NS_X_CONFERENCE, NULL))
-            break;
-
-      if (!node)
-        return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
+      node = lm_message_node_get_child_with_namespace (message->node, "x",
+          NS_X_CONFERENCE);
+      if (node == NULL)
+        goto HANDLE_MESSAGE;
 
       /* the room JID is in x */
       from = lm_message_node_get_attribute (node, "jid");
-      if (!from)
+      if (from == NULL)
         return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
 
       /* the inviter JID is in "from" */
-      inviter_handle = gabble_handle_for_contact (priv->conn->handles,
-                                                  from, FALSE);
+      inviter_handle = gabble_handle_for_contact (
+          conn->handles[TP_HANDLE_TYPE_CONTACT], from, FALSE);
 
       /* reason is the body */
       reason = body;
 
-      disco_udata = g_new0 (struct DiscoInviteData, 1);
+      disco_udata = g_slice_new0 (struct DiscoInviteData);
       disco_udata->factory = fac;
       disco_udata->reason = g_strdup (reason);
       disco_udata->inviter = inviter_handle;
 
       NODE_DEBUG (message->node, "received obsolete invite method");
-      /* FIXME should we keep track of these so we can cancel/invalidate
-       * them in dispose? */
 
       request = gabble_disco_request (priv->conn->disco, GABBLE_DISCO_TYPE_INFO,
           from, NULL, obsolete_invite_disco_cb, disco_udata, G_OBJECT (fac), NULL);
 
-      if (request)
+      if (request != NULL)
         g_hash_table_insert (priv->disco_requests, request, NULL);
 
       return LM_HANDLER_RESULT_REMOVE_MESSAGE;
   }
 
+HANDLE_MESSAGE:
+
   /* check if a room with the jid exists */
-  if (!gabble_handle_for_room_exists (priv->conn->handles, from, TRUE))
+  if (!gabble_handle_for_room_exists (
+        conn->handles[TP_HANDLE_TYPE_ROOM], from, TRUE))
     {
       return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
     }
 
-  room_handle = gabble_handle_for_room (priv->conn->handles, from);
+  room_handle = gabble_handle_for_room (
+      conn->handles[TP_HANDLE_TYPE_ROOM], from);
 
   /* find the MUC channel */
   chan = g_hash_table_lookup (priv->channels, GUINT_TO_POINTER (room_handle));
 
-  if (!chan)
+  if (chan == NULL)
     {
       g_warning ("%s: ignoring groupchat message from known handle with "
                  "no MUC channel", G_STRFUNC);
@@ -524,7 +540,8 @@ muc_factory_message_cb (LmMessageHandler *handler,
 
   /* get the handle of the sender, which is either the room
    * itself or one of its members */
-  if (gabble_handle_for_room_exists (priv->conn->handles, from, FALSE))
+  if (gabble_handle_for_room_exists (
+        conn->handles[TP_HANDLE_TYPE_ROOM], from, FALSE))
     {
       handle_type = TP_HANDLE_TYPE_ROOM;
       handle = room_handle;
@@ -532,16 +549,26 @@ muc_factory_message_cb (LmMessageHandler *handler,
   else
     {
       handle_type = TP_HANDLE_TYPE_CONTACT;
-      handle = gabble_handle_for_contact (priv->conn->handles, from, TRUE);
+      handle = gabble_handle_for_contact (
+          conn->handles[TP_HANDLE_TYPE_CONTACT], from, TRUE);
     }
 
-  if (send_error != CHANNEL_TEXT_SEND_NO_ERROR)
+  if (send_error != TP_CHANNEL_SEND_NO_ERROR)
     {
-      _gabble_text_mixin_send_error_signal (G_OBJECT (chan), send_error, stamp,
-          msgtype, body_offset);
+      tp_svc_channel_type_text_emit_send_error (
+          (TpSvcChannelTypeText *)chan, send_error, stamp, msgtype,
+          body_offset);
       return LM_HANDLER_RESULT_REMOVE_MESSAGE;
     }
 
+  if (state != -1) {
+    TpHandle from_handle;
+
+    from_handle = gabble_handle_for_contact (
+          priv->conn->parent.handles[TP_HANDLE_TYPE_CONTACT], from, FALSE);
+    _gabble_muc_channel_state_receive (chan, state, from_handle);
+  }
+
   if (_gabble_muc_channel_receive (chan, msgtype, handle_type, handle, stamp,
                                    body_offset, message))
     return LM_HANDLER_RESULT_REMOVE_MESSAGE;
@@ -567,6 +594,7 @@ muc_factory_presence_cb (LmMessageHandler *handler,
 {
   GabbleMucFactory *fac = GABBLE_MUC_FACTORY (user_data);
   GabbleMucFactoryPrivate *priv = GABBLE_MUC_FACTORY_GET_PRIVATE (fac);
+  TpBaseConnection *conn = (TpBaseConnection *)priv->conn;
   const char *from;
   LmMessageSubType sub_type;
   GabbleMucChannel *muc_chan;
@@ -598,13 +626,14 @@ muc_factory_presence_cb (LmMessageHandler *handler,
   x_node = lm_message_node_get_child_with_namespace (msg->node, "x", NS_MUC_USER);
 
   /* is it a MUC member presence? */
-  if (x_node)
+  if (x_node != NULL)
     {
       if (muc_chan != NULL)
         {
-          GabbleHandle handle;
+          TpHandle handle;
 
-          handle = gabble_handle_for_contact (priv->conn->handles, from, TRUE);
+          handle = gabble_handle_for_contact (
+              conn->handles[TP_HANDLE_TYPE_CONTACT], from, TRUE);
           if (handle == 0)
             {
               NODE_DEBUG (msg->node, "discarding MUC presence from malformed jid");
@@ -626,6 +655,54 @@ muc_factory_presence_cb (LmMessageHandler *handler,
 
 }
 
+static void
+roomlist_channel_closed_cb (GabbleRoomlistChannel *chan, gpointer data)
+{
+  GabbleMucFactory *fac = GABBLE_MUC_FACTORY (data);
+  GabbleMucFactoryPrivate *priv = GABBLE_MUC_FACTORY_GET_PRIVATE (fac);
+
+  if (priv->roomlist_channel != NULL)
+    {
+      g_object_unref (priv->roomlist_channel);
+      priv->roomlist_channel = NULL;
+    }
+}
+
+static gboolean
+make_roomlist_channel (GabbleMucFactory *fac, gpointer request)
+{
+  GabbleMucFactoryPrivate *priv = GABBLE_MUC_FACTORY_GET_PRIVATE (fac);
+  TpBaseConnection *conn = (TpBaseConnection *)priv->conn;
+
+  const gchar *server;
+  gchar *object_path;
+
+  g_assert (priv->roomlist_channel == NULL);
+
+  server = _gabble_connection_find_conference_server (priv->conn);
+
+  if (server == NULL)
+    return FALSE;
+
+  object_path = g_strdup_printf ("%s/RoomlistChannel",
+      conn->object_path);
+
+  priv->roomlist_channel = _gabble_roomlist_channel_new (priv->conn,
+      object_path, server);
+
+  g_signal_connect (priv->roomlist_channel, "closed",
+                    (GCallback) roomlist_channel_closed_cb, fac);
+
+  tp_channel_factory_iface_emit_new_channel (fac, 
+      (TpChannelIface *)priv->roomlist_channel, request);
+
+  g_free (object_path);
+
+  return TRUE;
+}
+
+
+
 static void
 gabble_muc_factory_iface_close_all (TpChannelFactoryIface *iface)
 {
@@ -634,12 +711,19 @@ gabble_muc_factory_iface_close_all (TpChannelFactoryIface *iface)
 
   DEBUG ("closing channels");
 
-  if (priv->channels)
+  if (priv->channels != NULL)
     {
       GHashTable *tmp = priv->channels;
       priv->channels = NULL;
       g_hash_table_destroy (tmp);
     }
+
+  if (priv->roomlist_channel != NULL)
+    {
+      GObject *tmp = G_OBJECT (priv->roomlist_channel);
+      priv->roomlist_channel = NULL;
+      g_object_unref (tmp);
+    }
 }
 
 static void
@@ -719,6 +803,9 @@ gabble_muc_factory_iface_foreach (TpChannelFactoryIface *iface, TpChannelFunc fo
   data.foreach = foreach;
 
   g_hash_table_foreach (priv->channels, _foreach_slave, &data);
+
+  if (priv->roomlist_channel != NULL)
+    foreach (TP_CHANNEL_IFACE (priv->roomlist_channel), user_data);
 }
 
 static TpChannelFactoryRequestStatus
@@ -726,14 +813,32 @@ gabble_muc_factory_iface_request (TpChannelFactoryIface *iface,
                                   const gchar *chan_type,
                                   TpHandleType handle_type,
                                   guint handle,
-                                  TpChannelIface **ret)
+                                  gpointer request,
+                                  TpChannelIface **ret,
+                                  GError **error)
 {
   GabbleMucFactory *fac = GABBLE_MUC_FACTORY (iface);
   GabbleMucFactoryPrivate *priv = GABBLE_MUC_FACTORY_GET_PRIVATE (fac);
+  TpBaseConnection *conn = (TpBaseConnection *)priv->conn;
   GabbleMucChannel *chan;
-  GArray *members;
-  GError *error;
-  gboolean retval;
+
+  if (!tp_strdiff (chan_type, TP_IFACE_CHANNEL_TYPE_ROOM_LIST))
+    {
+      if (priv->roomlist_channel != NULL)
+        {
+          *ret = TP_CHANNEL_IFACE (priv->roomlist_channel);
+          return TP_CHANNEL_FACTORY_REQUEST_STATUS_EXISTING;
+        }
+
+      /* FIXME - delay if services aren't discovered yet? */
+      if (!make_roomlist_channel (fac, request))
+        {
+          DEBUG ("no conference server available for roomlist request");
+          return TP_CHANNEL_FACTORY_REQUEST_STATUS_NOT_AVAILABLE;
+        }
+      *ret = TP_CHANNEL_IFACE (priv->roomlist_channel);
+      return TP_CHANNEL_FACTORY_REQUEST_STATUS_CREATED;
+    }
 
   if (strcmp (chan_type, TP_IFACE_CHANNEL_TYPE_TEXT))
     return TP_CHANNEL_FACTORY_REQUEST_STATUS_NOT_IMPLEMENTED;
@@ -741,41 +846,21 @@ gabble_muc_factory_iface_request (TpChannelFactoryIface *iface,
   if (handle_type != TP_HANDLE_TYPE_ROOM)
     return TP_CHANNEL_FACTORY_REQUEST_STATUS_NOT_AVAILABLE;
 
-  if (!gabble_handle_is_valid (priv->conn->handles, TP_HANDLE_TYPE_ROOM, handle, NULL))
+  if (!tp_handle_is_valid (conn->handles[TP_HANDLE_TYPE_ROOM],
+        handle, NULL))
     return TP_CHANNEL_FACTORY_REQUEST_STATUS_INVALID_HANDLE;
 
   chan = g_hash_table_lookup (priv->channels, GINT_TO_POINTER (handle));
   if (!chan)
     {
-      chan = new_muc_channel (fac, handle);
-
-      members = g_array_sized_new (FALSE, FALSE, sizeof (GabbleHandle), 1);
-      g_array_append_val (members, priv->conn->self_handle);
-
-      retval = gabble_group_mixin_add_members (G_OBJECT (chan), members, "", &error);
-
-      g_array_free (members, TRUE);
-
-      if (!retval)
-        {
-          GError *close_err;
-
-          if (!gabble_muc_channel_close (GABBLE_MUC_CHANNEL (chan), &close_err))
-            {
-              g_error_free (close_err);
-            }
-
-          DEBUG ("error while adding self to group mixin");
-          return TP_CHANNEL_FACTORY_REQUEST_STATUS_NOT_AVAILABLE;
-        }
-
+      chan = new_muc_channel (fac, handle, TRUE);
       return TP_CHANNEL_FACTORY_REQUEST_STATUS_QUEUED;
     }
 
   if (_gabble_muc_channel_is_ready (chan))
     {
       *ret = TP_CHANNEL_IFACE (chan);
-      return TP_CHANNEL_FACTORY_REQUEST_STATUS_DONE;
+      return TP_CHANNEL_FACTORY_REQUEST_STATUS_EXISTING;
     }
   else
     {
