@@ -22,6 +22,7 @@
 #define __GABBLE_NAMESPACES__H__
 
 #define NS_CAPS                 "http://jabber.org/protocol/caps"
+#define NS_CHAT_STATES          "http://jabber.org/protocol/chatstates"
 #define NS_DISCO_INFO           "http://jabber.org/protocol/disco#info"
 #define NS_DISCO_ITEMS          "http://jabber.org/protocol/disco#items"
 #define NS_GABBLE_CAPS          "http://telepathy.freedesktop.org/caps"
@@ -36,17 +37,20 @@
 #define NS_JINGLE               "http://jabber.org/protocol/jingle"
 #define NS_JINGLE_DESCRIPTION_AUDIO "http://jabber.org/protocol/jingle/description/audio"
 #define NS_JINGLE_DESCRIPTION_VIDEO "http://jabber.org/protocol/jingle/description/video"
+#define NS_JINGLE_ERRORS            "http://jabber.org/protocol/jingle#errors"
 #define NS_JINGLE_TRANSPORT_ICE     "http://jabber.org/protocol/jingle/transport/ice"
 #define NS_MUC                  "http://jabber.org/protocol/muc"
 #define NS_MUC_USER             "http://jabber.org/protocol/muc#user"
 #define NS_MUC_ADMIN            "http://jabber.org/protocol/muc#admin"
 #define NS_MUC_OWNER            "http://jabber.org/protocol/muc#owner"
 #define NS_NICK                 "http://jabber.org/protocol/nick"
+#define NS_NOKIA_IV             "http://videovoip.tableteer.nokia.com/xmpp/privilege"
 #define NS_PRESENCE_INVISIBLE   "presence-invisible"
 #define NS_PRIVACY              "jabber:iq:privacy"
 #define NS_REGISTER             "jabber:iq:register"
 #define NS_ROSTER               "jabber:iq:roster"
 #define NS_VCARD_TEMP           "vcard-temp"
+#define NS_VCARD_TEMP_UPDATE    "vcard-temp:x:update"
 #define NS_X_DATA               "jabber:x:data"
 #define NS_X_DELAY              "jabber:x:delay"
 #define NS_X_CONFERENCE         "jabber:x:conference"
