@@ -19,17 +19,18 @@
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
+#ifndef __GABBLE_UTIL_H__
+#define __GABBLE_UTIL_H__
+
 #include <glib.h>
+#include <telepathy-glib/util.h>
 #include <loudmouth/loudmouth.h>
 
 #include "gabble-types.h"
 
-#ifndef __GABBLE_UTIL_H__
-#define __GABBLE_UTIL_H__
-
-gboolean g_strdiff (const gchar *left, const gchar *right);
-void lm_message_node_add_nick (LmMessageNode *node, const gchar *nick);
+gchar *sha1_hex (const gchar *bytes, guint len);
 void lm_message_node_add_own_nick (LmMessageNode *node, GabbleConnection *conn);
+void lm_message_node_unlink (LmMessageNode *orphan);
 void lm_message_node_steal_children (LmMessageNode *snatcher, LmMessageNode *mum);
 gboolean lm_message_node_has_namespace (LmMessageNode *node, const gchar *ns, const gchar *tag);
 LmMessageNode *lm_message_node_get_child_with_namespace (LmMessageNode *node, const gchar *name, const gchar *ns);
